from typing import Union

import numpy as np
from scipy.stats import kurtosis, skew

from .time_unit import get_integer_time_unit

__all__ = ["get_annualized_kurtosis", "get_annualized_mean", "get_annualized_sd", "get_annualized_skew"]


def get_annualized_kurtosis(data: np.ndarray, time_unit: Union[int, str] = 'monthly') -> np.ndarray:
    """
    Gets the annualized standard deviation for each asset class in the data cube

    Parameters
    ----------
    data
        Data tensor. The axis must represent time, trials and assets respectively.

    time_unit: int or str
        Specifies how many units (first axis) is required to represent a year. For example, if each time period
        represents a month, set this to 12. If quarterly, set to 4. Defaults to 12 which means 1 period represents
        a month. Alternatively, you could put in a string name of the time_unit. Accepted values are weekly,
        monthly, quarterly, semi-annually and yearly

    Returns
    -------
    ndarray
        An array containing the annualized standard deviation for each asset class.
    """
    return np.mean([kurtosis(i) for i in annualize_data(data, time_unit)], 0)


def get_annualized_mean(data: np.ndarray, time_unit: Union[int, str] = 'monthly') -> np.ndarray:
    """
    Gets the annualized mean for each asset class in the data cube

    Parameters
    ----------
    data
        Data tensor. The axis must represent time, trials and assets respectively.

    time_unit: int or str
        Specifies how many units (first axis) is required to represent a year. For example, if each time period
        represents a month, set this to 12. If quarterly, set to 4. Defaults to 12 which means 1 period represents
        a month. Alternatively, you could put in a string name of the time_unit. Accepted values are weekly,
        monthly, quarterly, semi-annually and yearly

    Returns
    -------
    ndarray
        An array containing the annualized mean for each asset class.
    """
    time_unit = get_integer_time_unit(time_unit)
    years = len(data) // time_unit

    data = (annualize_data(data, time_unit) + 1).prod(0)
    return np.mean(np.sign(data) * np.abs(data) ** (1 / years), 0) - 1


def get_annualized_sd(data: np.ndarray, time_unit: Union[int, str] = 'monthly') -> np.ndarray:
    """
    Gets the annualized standard deviation for each asset class in the data cube

    Parameters
    ----------
    data
        Data tensor. The axis must represent time, trials and assets respectively.

    time_unit: int or str
        Specifies how many units (first axis) is required to represent a year. For example, if each time period
        represents a month, set this to 12. If quarterly, set to 4. Defaults to 12 which means 1 period represents
        a month. Alternatively, you could put in a string name of the time_unit. Accepted values are weekly,
        monthly, quarterly, semi-annually and yearly

    Returns
    -------
    ndarray
        An array containing the annualized standard deviation for each asset class.
    """
    return annualize_data(data, time_unit).std(1).mean(0)


def get_annualized_skew(data: np.ndarray, time_unit: Union[int, str] = 'monthly') -> np.ndarray:
    """
    Gets the annualized skew for each asset class in the data cube

    Parameters
    ----------
    data
        Data tensor. The axis must represent time, trials and assets respectively.

    time_unit: int or str
        Specifies how many units (first axis) is required to represent a year. For example, if each time period
        represents a month, set this to 12. If quarterly, set to 4. Defaults to 12 which means 1 period represents
        a month. Alternatively, you could put in a string name of the time_unit. Accepted values are weekly,
        monthly, quarterly, semi-annually and yearly

    Returns
    -------
    ndarray
        An array containing the annualized skew for each asset class.
    """
    return np.mean([skew(i) for i in annualize_data(data, time_unit)], 0)


def annualize_data(data: np.ndarray, time_unit: Union[str, int]) -> np.ndarray:
    data = np.asarray(data)
    time_unit = get_integer_time_unit(time_unit)

    t, n, a = data.shape

    return (data.reshape((t // time_unit, time_unit, n, a)) + 1).prod(1) - 1



from typing import Optional

import numpy as np
from arch.univariate.distribution import StudentsT as T

from ._base import DistributionMixin, _format_simulator


class StudentsT(T, DistributionMixin):
    def __init__(self, random_state=None):
        DistributionMixin.__init__(self)
        T.__init__(self, random_state)

    @_format_simulator
    def _simulator(self, size: int, reps: Optional[int] = None):
        nu = self._parameters[0]
        std_dev = np.sqrt(nu / (nu - 2))

        if self.custom_dist is None:
            if reps is not None:
                size = size, reps
            return self._random_state.standard_t(nu, size=size) / std_dev
        else:
            self.check_dist_size(size)
            ppf = self.ppf(self.custom_dist[:size], nu)
            self.custom_dist = None  # reset simulator

            return ppf / std_dev



from typing import Callable

import numpy as np
import pandas as pd
from arch.univariate.mean import ARX as A, ConstantMean as C, HARX as H, LS as _LS, ZeroMean as Z

from .simulators import harx_mc_simulator, harx_simulator

__all__ = ['ARX', 'ConstantMean', 'HARX', 'LS', 'ZeroMean']

GeneratorFunc = Callable[[int], np.ndarray]


class ARX(A):
    @harx_simulator
    def simulate(self, params, nobs, burn=500, initial_value=None, x=None, initial_value_vol=None, data_only=False):
        pass

    @harx_mc_simulator
    def simulate_mc(self, params, nobs, reps, burn=500, initial_value=None, x=None, initial_value_vol=None):
        pass


class HARX(H):
    @harx_simulator
    def simulate(self, params, nobs, burn=500, initial_value=None, x=None, initial_value_vol=None, data_only=False):
        pass

    @harx_mc_simulator
    def simulate_mc(self, params, nobs, reps, burn=500, initial_value=None, x=None, initial_value_vol=None):
        pass


class LS(_LS):
    @harx_simulator
    def simulate(self, params, nobs, burn=500, initial_value=None, x=None, initial_value_vol=None, data_only=False):
        """
        Examples
        --------
        Basic data simulation with a constant mean and volatility

        >>> import numpy as np
        >>> from muarch.mean import LS
        >>> from muarch.volatility import GARCH

        >>> ls = LS()
        >>> ls.volatility = GARCH()
        >>> ls_params = np.array([1, 1.5])
        >>> garch_params = np.array([0.01, 0.07, 0.92])
        >>> params = np.concatenate((ls_params, garch_params))
        >>> nobs, burn = 100, 400
        >>> sim_data = ls.simulate(params, nobs, burn,  x = np.random.normal(size=[nobs + burn, 1]))
        """
        pass

    @harx_mc_simulator
    def simulate_mc(self, params, nobs, reps, burn=500, initial_value=None, x=None, initial_value_vol=None):
        """
        Examples
        --------
        >>> import numpy as np
        >>> from muarch import MUArch
        >>> from muarch.datasets import load_etf

        >>> prices = load_etf()
        >>> CONST_GARCH11= MUArch(3, mean='Constant', lags=1, vol='GARCH', p=1, q=1, scale=100)
        >>> CONST_GARCH11.fit(prices)

        >>> sim_data = CONST_GARCH11.simulate_mc(nobs=36, reps=10, n_jobs=-1)
        """
        pass


class ConstantMean(C):
    def simulate(self, params, nobs, burn=500, initial_value=None, x=None, initial_value_vol=None, data_only=False):
        """
        Simulated data from a Constant Mean model

        Parameters
        ----------
        params: ndarray
            Parameters to use when simulating the model. Parameter order is [mean volatility distribution] where the
            parameters of the mean model are ordered [constant lag[0] lag[1] ... lag[p] ex[0] ... ex[k-1]] where lag[j]
            indicates the coefficient on the jth lag in the model and ex[j] is the coefficient on the jth exogenous
            variable.

        nobs: int
            Length of series to simulate

        burn: int, default 500
            Number of values to simulate to initialize the model and remove dependence on initial values

        initial_value: {ndarray, float}, optional
            Either a scalar value or `max(lags)` array set of initial values to use when initializing the model.
            If omitted, 0.0 is used

        x: {ndarray, DataFrame}, optional
            nobs + burn by k array of exogenous variables to include in the simulation. This should be a 2D matrix

        initial_value_vol: {ndarray, float}, optional
            An array or scalar to use when initializing the volatility process.

        data_only: bool, optional
            If True, this returns only the simulated data, omits the volatility and error. In this case, it will return
            as a numpy array. Otherwise, it returns a data frame with the data, volatility and error

        Returns
        -------
        simulated_data : {DataFrame, ndarray}
            DataFrame with columns data containing the simulated values, volatility, containing the conditional
            volatility and errors containing the errors used in the simulation.
            If data_only, it returns the 'data' column as a numpy array

        Examples
        --------
        Basic data simulation with a constant mean and volatility

        >>> import numpy as np
        >>> from muarch.mean import ConstantMean
        >>> from muarch.volatility import GARCH

        >>> cm = ConstantMean()
        >>> cm.volatility = GARCH()
        >>> cm_params = np.array([1])
        >>> garch_params = np.array([0.01, 0.07, 0.92])
        >>> params = np.concatenate((cm_params, garch_params))
        >>> sim_data = cm.simulate(params, 20)
        """
        if initial_value is not None or x is not None:
            raise ValueError('Both initial value and x must be none when '
                             'simulating a constant mean process.')

        mp, vp, dp = self._parse_parameters(params)

        sim_values = self.volatility.simulate(vp,
                                              nobs + burn,
                                              self.distribution.simulate(dp),
                                              burn,
                                              initial_value_vol)
        errors = sim_values[0]
        y = errors + mp

        if data_only:
            return y[burn:]

        vol = np.sqrt(sim_values[1])
        df = dict(data=y[burn:], volatility=vol[burn:], errors=errors[burn:])
        return pd.DataFrame(df)

    def simulate_mc(self, params, nobs, reps, burn=500, initial_value=None, x=None, initial_value_vol=None):
        """
        Monte Carlo simulation from a Constant Mean model

        Parameters
        ----------
        params: ndarray
            Parameters to use when simulating the model. Parameter order is
            [mean volatility distribution] where the parameters of the mean
            model are ordered [constant lag[0] lag[1] ... lag[p] ex[0] ...
            ex[k-1]] where lag[j] indicates the coefficient on the jth lag in
            the model and ex[j] is the coefficient on the jth exogenous
            variable.

        nobs: int
            Length of series to simulate

        reps: int
            Number of trials in the Monte Carlo simulation

        burn: int, default 500
            Number of values to simulate to initialize the model and remove dependence on initial values

        initial_value: {ndarray, float}, optional
            Either a scalar value or `max(lags)` array set of initial values to use when initializing the model.
            If omitted, 0.0 is used

        x: {ndarray, DataFrame}, optional
            nobs + burn by k array of exogenous variables to include in the simulation. This should be a 2D matrix

        initial_value_vol: {ndarray, float}, optional
            An array or scalar to use when initializing the volatility process.

        Returns
        -------
        simulated_data: ndarray
            matrix of simulated values

        Examples
        --------
        >>> import numpy as np
        >>> from muarch import MUArch
        >>> from muarch.datasets import load_etf

        >>> prices = load_etf()
        >>> CONST_GARCH11= MUArch(3, mean='Constant', lags=1, vol='GARCH', p=1, q=1, scale=100)
        >>> CONST_GARCH11.fit(prices)
        >>> sim_data = CONST_GARCH11.simulate_mc(nobs=36, reps=10, n_jobs=-1)
        """
        if initial_value is not None or x is not None:
            raise ValueError('Both initial value and x must be none when '
                             'simulating a constant mean process.')

        mp, vp, dp = self._parse_parameters(params)
        simulator = self.distribution.simulate(dp)
        errors = self.volatility.simulate_mc(vp, nobs + burn, reps, simulator, burn, initial_value_vol)
        y = errors + mp

        return y[burn:]


class ZeroMean(Z):
    def simulate(self, params, nobs, burn=500, initial_value=None, x=None, initial_value_vol=None, data_only=False):
        """
        Simulated data from a Zero Mean model

        params: ndarray
            Parameters to use when simulating the model. Parameter order is
            [mean volatility distribution] where the parameters of the mean
            model are ordered [constant lag[0] lag[1] ... lag[p] ex[0] ...
            ex[k-1]] where lag[j] indicates the coefficient on the jth lag in
            the model and ex[j] is the coefficient on the jth exogenous
            variable.

        nobs: int
            Length of series to simulate

        burn: int, optional
            Number of values to simulate to initialize the model and remove dependence on initial values

        initial_value: {ndarray, float}, optional
            Either a scalar value or `max(lags)` array set of initial values to use when initializing the model.
            If omitted, 0.0 is used

        x: {ndarray, DataFrame}, optional
            nobs + burn by k array of exogenous variables to include in the simulation. This should be a 2D matrix

        initial_value_vol: {ndarray, float}, optional
            An array or scalar to use when initializing the volatility process.

        data_only: bool, optional
            If True, this returns only the simulated data, omits the volatility and error. In this case, it will return
            as a numpy array. Otherwise, it returns a data frame with the data, volatility and error

        Returns
        -------
        simulated_data: {DataFrame ,ndarray}
            DataFrame with columns data containing the simulated values, volatility, containing the conditional
            volatility and errors containing the errors used in the simulation.
            If data_only, it returns the 'data' column as a numpy array

        Examples
        --------
        Basic data simulation with no mean and constant volatility

        >>> from muarch.mean import ZeroMean
        >>> zm = ZeroMean()
        >>> sim_data = zm.simulate([1.0], 20)

        Simulating data with a non-trivial volatility process

        >>> from muarch.volatility import GARCH
        >>> zm.volatility = GARCH(p=1, o=1, q=1)
        >>> sim_data = zm.simulate([0.05, 0.1, 0.1, 0.8], 300)
        """
        if initial_value is not None or x is not None:
            raise ValueError('Both initial value and x must be none when simulating a constant mean process.')

        _, vp, dp = self._parse_parameters(params)

        sim_values = self.volatility.simulate(vp,
                                              nobs + burn,
                                              self.distribution.simulate(dp),
                                              burn,
                                              initial_value_vol)
        errors = sim_values[0]
        y = errors

        if data_only:
            return np.asarray(y[burn:])

        vol = np.sqrt(sim_values[1])
        df = dict(data=y[burn:], volatility=vol[burn:], errors=errors[burn:])
        return pd.DataFrame(df)

    def simulate_mc(self, params, nobs, reps, burn=500, initial_value=None, x=None, initial_value_vol=None):
        """
        Monte Carlo simulation from a constant mean model

        Parameters
        ----------
        params: ndarray
            Parameters to use when simulating the model. Parameter order is
            [mean volatility distribution] where the parameters of the mean
            model are ordered [constant lag[0] lag[1] ... lag[p] ex[0] ...
            ex[k-1]] where lag[j] indicates the coefficient on the jth lag in
            the model and ex[j] is the coefficient on the jth exogenous
            variable.

        nobs: int
            Length of series to simulate

        reps: int
            Number of trials in the Monte Carlo simulation

        burn: int, optional
            Number of values to simulate to initialize the model and remove dependence on initial values

        initial_value: {ndarray, float}, optional
            Either a scalar value or `max(lags)` array set of initial values to use when initializing the model.
            If omitted, 0.0 is used

        x: {ndarray, DataFrame}, optional
            nobs + burn by k array of exogenous variables to include in the simulation. This should be a 2D matrix

        initial_value_vol: {ndarray, float}, optional
            An array or scalar to use when initializing the volatility process.

        Returns
        -------
        simulated_data: ndarray
            matrix of simulated values

        Examples
        --------
        >>> import numpy as np
        >>> from muarch import MUArch
        >>> from muarch.datasets import load_etf

        >>> prices = load_etf()
        >>> CONST_GARCH11= MUArch(3, mean='Constant', lags=1, vol='GARCH', p=1, q=1, scale=100)
        >>> CONST_GARCH11.fit(prices)
        >>> sim_data = CONST_GARCH11.simulate_mc(nobs=36, reps=10, n_jobs=-1)
        """
        if initial_value is not None or x is not None:
            raise ValueError('Both initial value and x must be none when simulating a constant mean process.')

        _, vp, dp = self._parse_parameters(params)

        simulator = self.distribution.simulate(dp)
        errors = self.volatility.simulate_mc(vp, nobs + burn, reps, simulator, burn, initial_value_vol)
        return errors[burn:]



from typing import Callable, Iterable, Optional, Union

import numpy as np
from arch.univariate.base import ARCHModelResult

from muarch.mean import ARX, ConstantMean, HARX, LS, ZeroMean

__all__ = ['_ArchBase', 'PARAMS', 'RNG_GEN']

RNG_GEN = Callable[[Union[int, Iterable[int]]], np.ndarray]
PARAMS = Optional[np.ndarray]


class _ArchBase:
    def __init__(self):
        self._model: Union[HARX, ARX, ConstantMean, LS, ZeroMean] = None
        self._fit_model: ARCHModelResult = None

    def simulation_horizon_required(self, nobs: int, burn: int):
        """
        Calculates the number of random generations needed for simulation

        Parameters
        ----------
        nobs: int
            number of observations

        burn: int
            number of observations burnt in simulation

        Returns
        -------
        int
            number of random generations required
        """
        num = nobs + burn * 2
        if self._model.distribution.__class__.__name__ == 'GeneralizedError':
            num *= 2
        return num



import numpy as np
import scipy.optimize as opt

from ._calibrate_utils import calc_current_sd, get_data_shape, validate_target_sd


def calibrate_sd_only(data: np.ndarray, sd: np.ndarray, time_unit: int):
    validate_target_sd(data, sd)
    curr_sd = calc_current_sd(data, time_unit)

    sol = get_solutions(data, time_unit, sd)
    for i, s in enumerate(sol):
        if np.isfinite(s) and s > 0:
            data[..., i] *= s
        else:
            data[..., i] *= sd[i] / curr_sd[i]

    return data


def get_solutions(data: np.ndarray, time_unit: int, sd: np.ndarray):
    asset_mean_list = (AssetSD(data[..., i], time_unit, target) for i, target in enumerate(sd))

    return [f() for f in asset_mean_list]


class AssetSD:
    def __init__(self, data: np.ndarray, time_unit, target_sd: float):
        assert data.ndim == 2
        self.data = data
        self.time_unit = time_unit
        self.years, self.trials = get_data_shape(data, time_unit)
        self.target = target_sd
        self.best_guess = self.calc_best_guess()

    def __call__(self) -> float:
        a, b = self.best_guess
        return opt.toms748(self.annualized_sd, a, b)

    def annualized_sd(self, x):
        shape = self.years, self.time_unit, self.trials
        return ((self.data * x + 1).reshape(shape).prod(1) - 1).std(1).mean() - self.target

    def calc_best_guess(self):
        space = 2 ** np.linspace(-15, 6, 30)
        f_space = np.array([self.annualized_sd(x) for x in space])
        mask = (f_space[:-1] * f_space[1:]) <= 0

        if not np.any(mask):
            raise RuntimeError("Unable to find roots")

        index = np.argmin(np.abs(f_space[:-1] - f_space[1:])[mask])  # index with best root character
        return space[:-1][mask][index], space[1:][mask][index]



from typing import Optional

import numpy as np
import scipy.special as sp
from arch.univariate.distribution import GeneralizedError as GE
from scipy.stats import gamma

from ._base import DistributionMixin, _format_simulator


class GeneralizedError(GE, DistributionMixin):
    def __init__(self, random_state=None):
        DistributionMixin.__init__(self)
        GE.__init__(self, random_state)

    @_format_simulator
    def _simulator(self, size: int, reps: Optional[int] = None):
        _parameters = self._parameters
        if self.custom_dist is None:
            if reps is not None:
                size = size, reps

            nu, *_ = _parameters
            randoms = self._random_state.standard_gamma(1 / nu, size) ** (1.0 / nu)
            randoms *= 2 * self._random_state.randint(0, 2, size) - 1
            scale = np.sqrt(sp.gamma(3.0 / nu) / sp.gamma(1.0 / nu))

            return randoms / scale
        else:
            self.check_dist_size(size * 2)

            nu, *_ = _parameters

            randoms = gamma.ppf(self.custom_dist[:size], 1 / nu) ** (1. / nu)
            randoms *= 2 * np.asarray(self.custom_dist[size:2 * size] > 0.5, np.float) - 1
            scale = np.sqrt(sp.gamma(3.0 / nu) / sp.gamma(1.0 / nu))

            self.custom_dist = None  # reset simulator

            return randoms / scale



from copy import deepcopy
from typing import Iterable, Optional, Union

import numpy as np

from muarch.funcs.moments import get_annualized_mean, get_annualized_sd
from muarch.funcs.time_unit import get_integer_time_unit
from ._calibrate_both import calibrate_mean_and_sd
from ._calibrate_mean import calibrate_mean_only
from ._calibrate_sd import calibrate_sd_only


def calibrate_data(data: np.ndarray, mean: Optional[Iterable[float]] = None, sd: Optional[Iterable[float]] = None,
                   time_unit: Union[int, str] = "month", inplace=False, tol=6) -> np.ndarray:
    """
    Calibrates the data given the target mean and standard deviation.

    Parameters
    ----------
    data: ndarray
        Data tensor to calibrate

    mean: iterable float, optional
        The target annual mean vector

    sd: iterable float, optional
        The target annual standard deviation (volatility) vector

    time_unit: int or str
        Specifies how many units (first axis) is required to represent a year. For example, if each time period
        represents a month, set this to 12. If quarterly, set to 4. Defaults to 12 which means 1 period represents
        a month. Alternatively, you could put in a string name of the time_unit. Accepted values are weekly,
        monthly, quarterly, semi-annually and yearly

    inplace: bool
        If True, calibration will modify the original data. Otherwise, a deep copy of the original data will be
        made before calibration. Deep copy can be time consuming if data is big.

    tol: int
        Tolerance used to determine if calibrate should be called. For example, if the cube's target annualized
        mean is similar to the actual tolerance, function will skip the mean adjustment.

    Returns
    -------
    ndarray
        An instance of the adjusted numpy tensor
    """
    assert not np.isnan(data).any(), "data cube must not have nan values"
    time_unit = get_integer_time_unit(time_unit)

    if not inplace:
        data = deepcopy(data)

    mean = _set_to_none_if_close(mean, get_annualized_mean(data, time_unit))
    sd = _set_to_none_if_close(sd, get_annualized_sd(data, time_unit))

    return _calibrate(data, mean, sd, time_unit)


def _set_to_none_if_close(actual: Optional[np.ndarray], target: np.ndarray, tol=6):
    return None if actual is None or np.isclose(actual, target, atol=tol).all() else np.asarray(actual)


def _calibrate(data: np.ndarray, mean: Optional[np.ndarray], sd: Optional[np.ndarray], time_unit: int):
    if mean is not None and sd is not None:
        return calibrate_mean_and_sd(data, np.asarray(mean), np.asarray(sd), time_unit)

    if mean is not None and sd is None:
        return calibrate_mean_only(data, np.asarray(mean), time_unit)

    if mean is None and sd is not None:
        return calibrate_sd_only(data, np.asarray(sd), time_unit)

    return data  # no adjustments



import os
from collections import abc
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from typing import Iterable, List, Optional, Union

import numpy as np
import pandas as pd

from ._base import RNG_GEN
from .summary import SummaryList
from .uarch import UArch

Exog = Optional[Union[List[Union[None, np.ndarray]], np.ndarray]]


class MUArch:
    """
    Multi-univariate ARCH model. Unlike a multivariate ARCH model, this fits each univariate time series individually.
    Any simulations returns simulations of each univariate series column bound together.
    """

    __n: int
    __models: List[UArch]

    def __init__(self, n: Union[int, Iterable[UArch]], mean='Constant', lags=0, vol='GARCH', p=1, o=0, q=1, power=2.0,
                 dist='Normal', hold_back=None, scale=1):
        r"""
        Initializes the MUArch model.

        The MUArch model holds multiple univariate models which are determined during fitting. If the models are not
        specified, the global default options will be used. Models can be individually specified after initializing
        the MUArch instance.

        Parameters
        ----------
        n:  int or list of UArch models
            Number of univariate models to fit. Alternatively, a list of UArch (univariate) models can be specified.

        mean: { 'zero', 'constant', 'harx', 'har', 'ar', 'arx', 'ls' }, optional
            Name of the global default mean model.  Currently supported options are:

            * **Constant** (default) - Constant mean model
            * **Zero** - Zero mean model
            * **AR** - Autoregression model
            * **ARX** - Autoregression model with exogenous regressors. Falls back to **AR** if no exogenous regressors
            * **HAR** - Heterogeneous Autoregression model
            * **HARX** - Heterogeneous Autoregressions with exogenous regressors
            * **LS** - Least squares model

            For more information on the different models, check out the documentation at
            https://arch.readthedocs.io/en/latest/univariate/mean.html

        lags: int or list (int), optional
            Global default lag. Either a scalar integer value indicating lag length or a list of integers specifying
            lag locations.

        vol: { 'GARCH', 'ARCH', 'EGARCH', 'FIGARCH' and 'HARCH', 'CONSTANT' }, optional
            Name of the global default volatility model.  Currently supported options are:

            * **GARCH** (default) - Standard GARCH process which can be used to specify the following models:

                * ARCH(p)
                * GARCH(p,q)
                * GJR-GARCH(p,o,q)
                * AVARCH(p)
                * AVGARCH(p,q)
                * TARCH(p,o,q)
                * Models with arbitrary, pre-specified powers

            * **ARCH** - ARCH process
            * **EGARCH** - EGARCH process
            * **FIGARCH** - Fractionally Integrated (FI) GARCH process
            * **HARCH** - Heterogeneous ARCH process
            * **Constant** (default) - Constant volatility process

        p: int, optional
            Global default lag order of the symmetric innovation

        o: int, optional
            Global default lag order of the asymmetric innovation

        q: int, optional
            Global default lag order of lagged volatility or equivalent

        power: float, optional
            Global default power to use with the innovations, Default is 2.0, which produces ARCH and related
            models. Using 1.0 produces AVARCH and related models. Other powers can be specified, although these should
            be strictly positive, and usually larger than 0.25.

        dist:  { 'normal', 'gaussian', 'studentst', 't', 'skewstudent', 'skewt', 'ged', 'generalized error' }, optional
            Name of the global default distribution for the innovations.  Currently supported options are:

            * **normal**, **gaussian** (default) - Standard Normal distribution
            * **t**, **studentst** - Standardized Student’s distribution
            * **skewstudent**, **skewt** - Standardized Skewed Student’s distribution.
            * **ged**, **generalized error" - Generalized Error Distribution

        hold_back: int
            Global default. Number of observations at the start of the sample to exclude when estimating model
            parameters. Used when comparing models with different lag lengths to estimate on the common sample.

        scale: float
            Global default factor to scale data up or down by. This is useful when your data is too small leading to
            numerical errors when fitting. It will be used to scale simulation data
        """
        super(MUArch, self).__init__()

        if isinstance(n, abc.Iterable):
            self.__models = list(n)
            self.__n = len(self.__models)

            assert self.__n != 0, "If passing in a list of UArch models, list cannot be empty!"

            for i, m in enumerate(self.__models):
                assert isinstance(m, UArch), f'The model in index {i} must be an instance of UArch'

            self._model_names = [str(i) for i in range(self.__n)]

        elif type(n) is int:
            self.__n = n
            self.__models = [UArch(mean, lags, vol, p, o, q, power, dist, hold_back, scale) for _ in range(n)]
            self._model_names = [str(i) for i in range(n)]
        else:
            raise TypeError('`n` must either be an integer specifying the number of models or a list of UArch models')

    def fit(self, y: Union[pd.DataFrame, np.ndarray], x: Exog = None, update_freq=1, disp='off', cov_type='robust',
            show_warning=True, tol=None, options=None):
        """
        Fits the MUArch model.

        If finer control over the MUArch models is required, set the UArch models separately. Otherwise, method will
        set the default parameters.

        Parameters
        ----------
        y: {ndarray, Series}
            The dependent variable. If a vector is passed in, it is assumed that the same vector (endog) is used for
            all models. Otherwise, the last value of the shape must match the number of models

        x: list of {ndarray, None}, optional
            Exogenous regressors.  Ignored if model does not permit exogenous regressors. If passed in, the first shape
            must match the number of models.

        update_freq: int, optional
            Frequency of iteration updates.  Output is generated every `update_freq` iterations. Set to 0 to disable
            iterative output

        disp: 'final' or 'off' (default)
            Either 'final' to print optimization result or 'off' to display nothing

        cov_type: str, optional
            Estimation method of parameter covariance.  Supported options are 'robust', which does not assume the
            Information Matrix Equality holds and 'classic' which does. In the ARCH literature, 'robust' corresponds
            to Bollerslev-Wooldridge covariance estimator.

        show_warning: bool, optional
            Flag indicating whether convergence warnings should be shown.

        tol: float, optional
            Tolerance for termination

        options: dict, optional
            Options to pass to `scipy.optimize.minimize`.  Valid entries include 'ftol', 'eps', 'disp', and 'maxiter'

        Returns
        -------
        MUArch
            Fitted self instance
        """

        if y.shape[1] != self.__n:  # checks on input data
            raise ValueError("number of columns in data 'y' does not match expected dimension of MUArch object")

        if isinstance(y, pd.DataFrame):  # setting names for params later
            self._model_names = list(y.columns)
            y = y.values
        else:
            self._model_names = [str(i) for i in range(self.__n)]

        if y.ndim == 1:
            y = np.tile(y, self.__n)
        elif y.ndim > 2:
            raise ValueError("Dependent variable `y` must either be 1 or 2 dimensional")

        for i in range(self.__n):
            yy = y[:, i]
            xx = x[i] if x is not None else None
            if xx is not None:
                xx = np.asarray(xx)
                if xx.ndim == 1:
                    xx = xx.reshape(len(xx), 1)

            self[i].fit(yy, xx, update_freq=update_freq, disp=disp, cov_type=cov_type, show_warning=show_warning,
                        tol=tol,
                        options=options)

    @property
    def params(self):
        return pd.DataFrame([m.params for m in self], self._model_names)

    def residuals(self, standardize=True) -> np.ndarray:
        """
        Model residuals

        The residuals will be burnt by the maximum lag of the underlying models. For example, given 3 models - AR(1),
        AR(10), Constant with 400 data points each, the residuals will be 399, 390 and 400 long. The function will
        cut off the first 10 data points in this instance.

        Parameters
        ----------
        standardize: bool, optional
            Whether to standardize residuals. Residuals are standardized by dividing it with the conditional volatility

        Returns
        -------
        ndarray
            Residuals
        """
        residuals = [mm.residuals(standardize) for mm in self]
        length = min([len(r) for r in residuals])

        return np.asarray([r[-length:] for r in residuals]).T

    def simulate(self,
                 nobs,
                 burn=500,
                 initial_value=None,
                 x=None,
                 initial_value_vol=None,
                 data_only=True,
                 custom_dist: Optional[Union[RNG_GEN, np.ndarray]] = None):
        """
        Simulates data from the multiple ARMA-GARCH models

        Parameters
        ----------
        nobs: int
            Length of series to simulate

        burn: int, optional
            Number of values to simulate to initialize the model and remove dependence on initial values

        initial_value: {ndarray, float}, optional
            Either a scalar value or `max(lags)` array set of initial values to use when initializing the model.
            If omitted, 0.0 is used. If array, the last column must be of the same size as the number of models

        x: {ndarray, list of ndarray}, optional
            If supplied as a list, this list should have the same number of elements as the number of models in the
            MUArch model. Each array inside is the specified exogenous variable for that  particular model and
            this must be a nobs + burn by k matrix of exogenous variables to include in the simulation. Otherwise,
            leave the value as `None` to indicate no exogenous variables are used for simulation in the model.

            If an array is supplied directly, it means every model has an exogenous variable associated with it. In
            this case, it should be a 3 dimensional tensor where the first dimension represents the number of models.

        initial_value_vol: {ndarray, float}, optional
            An array or scalar to use when initializing the volatility process. If array, the last column must be of the
            same size as the number of models

        data_only: bool, default True
            If True, this returns only the simulated data, omits the volatility and error. In this case, it will return
            as a numpy array. Otherwise, it returns a data frame with the data, volatility and error

        custom_dist: {ndarray, Callable}, optional
            Optional density from which to simulate the innovations (Distribution) in the GARCH models. This is useful
            when working with the copula-GARCH model where each univariate model innovations has dependence on others.
            It is assumed that the values supplied are standardized [0, 1] innovations instead of the unstandardized
            residuals.

            The shape of the array must be at least as long as the simulation size required after accounting for burn
            and type of innovation process. If unsure, use :code:`simulation_size_required` to check. It must also
            account for the number of dimensions of the MUArch model. For example, if MUArch model is simulating a
            horizon of 120 time steps, 10000 trials and has 5 UArch models, the shape of the numpy array should be
            (120, 10000, 5).

            If a random number generator function is passed in, ensure that it only takes only argument and returns
            a numpy array. The argument can be an integer or a tuple of integers. In this case, the size will be
            automatically derived to save the user the trouble.

        Returns
        -------
        simulated_data : {List[DataFrame], ndarray}
            List of DataFrame with columns data containing the simulated values, volatility, containing the conditional
            volatility and errors containing the errors used in the simulation

        See Also
        --------
        UArch.simulation_horizon_required: Calculates the simulation size required
        """

        initial_value, initial_value_vol, x, custom_dist = self._format_simulation_parameters(nobs,
                                                                                              burn,
                                                                                              None,
                                                                                              initial_value,
                                                                                              initial_value_vol,
                                                                                              x,
                                                                                              custom_dist)

        sims = []
        for i in range(self.__n):
            r = custom_dist[..., i] if custom_dist is not None else None
            xx = x[i]
            iv = initial_value[i]
            ivv = initial_value_vol[i]

            sims.append(self[i].simulate(nobs, burn, iv, xx, ivv, data_only, None, r))

        if data_only:
            sims = np.asarray(sims).T

        return sims

    def simulate_mc(self,
                    nobs,
                    reps,
                    burn=500,
                    initial_value=None,
                    x=None,
                    initial_value_vol=None,
                    custom_dist: Optional[Union[RNG_GEN, np.ndarray]] = None,
                    n_jobs: Optional[int] = None):
        """
        Simulates data from the multiple ARCH-GARCH models.

        This function is specially crafted for Monte-Carlo simulations.

        Parameters
        ----------
        nobs: int
            Length of series to simulate

        reps: int
            Number of repetitions

        burn: int, optional
            Number of values to simulate to initialize the model and remove dependence on initial values

        initial_value: {ndarray, float}, optional
            Either a scalar value or `max(lags)` array set of initial values to use when initializing the model.
            If omitted, 0.0 is used. If array, the last column must be of the same size as the number of models

        x: {ndarray, list of ndarray}, optional
            If supplied as a list, this list should have the same number of elements as the number of models in the
            MUArch model. Each array inside is the specified exogenous variable for that  particular model and
            this must be a nobs + burn by k matrix of exogenous variables to include in the simulation. Otherwise,
            leave the value as `None` to indicate no exogenous variables are used for simulation in the model.

            If an array is supplied directly, it means every model has an exogenous variable associated with it. In
            this case, it should be a 3 dimensional tensor where the first dimension represents the number of models.

        initial_value_vol: {ndarray, float}, optional
            An array or scalar to use when initializing the volatility process. If array, the last column must be of the
            same size as the number of models

        custom_dist: {ndarray, Callable}, optional
            Optional density from which to simulate the innovations (Distribution) in the GARCH models. This is useful
            when working with the copula-GARCH model where each univariate model innovations has dependence on others.
            It is assumed that the values supplied are standardized [0, 1] innovations instead of the unstandardized
            residuals.

            The shape of the array must be at least as long as the simulation size required after accounting for burn
            and type of innovation process. If unsure, use :code:`simulation_size_required` to check.

            If a random number generator function is passed in, he size will be automatically derived to save the user
            the trouble. However, the function must:

            * take as it first argument an integer or a tuple of integer
            * have other parameters that are optional
            * return a numpy array

        n_jobs: int or None, optional
            The number of jobs to run in parallel for simulation. This is particularly useful when simulating large
            number of repetitions with more than 1 dimension. None defaults to using 1 processor. Any numbers less or
            equal to 0 means to use all processors. Even if a large number is used, it will be capped at the
            maximum number of processors available.


        Returns
        -------
        simulated_data : numpy array
            Array containing simulated data from the Monte Carlo Simulation
        """

        if not isinstance(reps, int) or reps < 1:
            raise ValueError('reps must be an integer greater than 0')

        # setting up job processes
        if n_jobs is None:
            n_jobs = 1
        elif not isinstance(n_jobs, int):
            raise ValueError('`n_jobs` must be an integer')
        else:
            n_jobs = min(os.cpu_count(), os.cpu_count() if n_jobs <= 0 else n_jobs)

        initial_value, initial_value_vol, x, custom_dist = self._format_simulation_parameters(nobs,
                                                                                              burn,
                                                                                              reps,
                                                                                              initial_value,
                                                                                              initial_value_vol,
                                                                                              x,
                                                                                              custom_dist)

        # formulate function calls to get simulated data
        functions = []
        for i in range(self.__n):
            r = custom_dist[..., i] if custom_dist is not None else None
            xx = x[i]
            iv = initial_value[i]
            ivv = initial_value_vol[i]

            func = partial(self[i].simulate_mc, nobs, reps, burn, iv, xx, ivv, None, r)
            functions.append(func)

        # switch between sequential and multiprocessing
        if n_jobs == 1:  # sequential
            sims = np.zeros((nobs, reps, self.__n))
            for i, func in enumerate(functions):
                sims[:, :, i] = func()

            return sims
        else:  # multiprocessing
            results = []

            with ProcessPoolExecutor(n_jobs) as P:
                for func in functions:
                    results.append(P.submit(func))

            results = np.asarray([r.result() for r in results])
            return np.moveaxis(results, 0, -1)

    def summary(self, short=False, dp=4):
        """
        Summary of fitted models

        :param short: bool, default False
            Whether to show short summary or full summary.
        :param dp: int, default 4
            Number of decimal places to show in short summary
        :return: SummaryList
            summary of fitted models
        """
        if short:
            return pd.DataFrame([m.summary(short, dp) for m in self], self._model_names)
        else:
            smry = SummaryList()
            for i, name in enumerate(self._model_names):
                s = self[i].summary(short, dp)
                s.add_header(name)
                smry.append(s)
            return smry

    def _format_simulation_parameters(self,
                                      nobs,
                                      burn,
                                      reps,
                                      initial_value,
                                      initial_value_vol,
                                      x,
                                      custom_dist):
        """Ensures the simulation are okay and return the formatted parameters"""

        # Format initial value
        if initial_value is None or isinstance(initial_value, (float, int)):
            initial_value = np.asarray([initial_value] * self.__n)

        # Format initial volatility value
        if initial_value_vol is None or isinstance(initial_value, (float, int)):
            initial_value_vol = np.asarray([initial_value_vol] * self.__n)

        # Format exogenous regressors
        if x is not None:
            if isinstance(x, np.ndarray):
                assert x.ndim == 3, ("If numpy array passed in as MUArch `x` (exog), make sure it is a 3 "
                                     "dimensional array where the last dimension is the number of elements")
                x = [x[i] for i in range(self.__n)]
            else:
                x = list(x)

                assert len(x) == self.__n, ("Exogenous variable's list should have the same number of elements as the "
                                            "number of models")

                for i, xx in enumerate(x):
                    if xx is None:
                        continue
                    elif isinstance(x[0], pd.DataFrame):
                        # convert all data frames to numpy array then move the model axis (first) to last
                        xx = np.moveaxis(np.array([df.values for df in x]), 0, -1)
                    else:
                        xx = np.asarray(xx)

                    if xx.ndim == 1:
                        xx = xx.reshape(-1, 1)
                    x[i] = xx
        else:
            x = [None] * self.__n

        # Format custom distribution
        if custom_dist is not None:
            horizon = max(m.simulation_horizon_required(nobs, burn) for m in self)
            if isinstance(reps, int):  # for monte carlo
                size = horizon, reps
            else:
                size = horizon

            if callable(custom_dist):
                custom_dist = np.asarray(custom_dist(size))
            else:
                custom_dist = np.asarray(custom_dist)

            if custom_dist.ndim == 1:
                custom_dist = custom_dist.reshape(1, -1)

            dim = custom_dist.shape[-1]
            assert dim == self.__n, (f"expected generator to return array with {self.__n} vectors in the last axis "
                                     f"(of shape) but got {dim}")

        return initial_value, initial_value_vol, x, custom_dist

    def _repr_html_(self):
        header = '<h2>Arch Models</h2>'
        models = '<br/>'.join(m._repr_html_() for m in self)
        return header + models

    def __getitem__(self, i: int):
        return self.__models[i]

    def __setitem__(self, i: int, model: UArch):
        assert isinstance(model, UArch), 'only UArch models are allowed to be set as a component of the MUArch object'
        self.__models[i] = model

    def __repr__(self):
        return self.__str__() + f', {hex(id(self))}'

    def __str__(self):
        txt = f'{"Arch Models":^80}\n' + '=' * 80 + '\n\n'
        txt += '\n\n\n'.join(str(m) for m in self)
        return txt

    def __len__(self):
        return self.__n



import numpy as np
import scipy.optimize as opt

from ._calibrate_utils import get_data_shape, validate_target_mean


def calibrate_mean_only(data: np.ndarray, mean: np.ndarray, time_unit: int):
    validate_target_mean(data, mean)

    sol = get_solutions(data, time_unit, mean)
    for i, s in enumerate(sol):
        data[..., i] += s

    return data


def get_solutions(data: np.ndarray, time_unit: int, mean: np.ndarray):
    asset_list = (AssetMean(data[..., i], time_unit, target) for i, target in enumerate(mean))

    return [f() for f in asset_list]


class AssetMean:
    def __init__(self, data: np.ndarray, time_unit, target_mean: float):
        assert data.ndim == 2
        self.data = data
        self.time_unit = time_unit
        self.years, _ = get_data_shape(data, time_unit)
        self.target = target_mean
        self.best_guess = self.calc_best_guess()

    def __call__(self) -> float:
        a, b = self.best_guess
        return opt.toms748(self.annualized_mean, a, b)

    def annualized_mean(self, x):
        d = (self.data + x + 1).prod(0)
        return (np.sign(d) * np.abs(d) ** (1 / self.years)).mean() - 1 - self.target

    def calc_best_guess(self):
        space = 2 ** np.linspace(-15, 4, 20)
        space = np.sort([*-space, *space])

        f_space = np.array([self.annualized_mean(x) for x in space])
        mask = (f_space[:-1] * f_space[1:]) <= 0

        if not np.any(mask):
            raise RuntimeError("Unable to find roots")

        index = np.argmin(np.abs(f_space[:-1] - f_space[1:])[mask])  # index with best root character
        return space[:-1][mask][index], space[1:][mask][index]



cimport cython

from libc.math cimport sqrt, pi
from libc.stdint cimport int64_t
import numpy as np



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def garch_simulate(int64_t p,
                   int64_t o,
                   int64_t q,
                   double power,
                   double[:] parameters,
                   int64_t nobs,
                   int64_t burn,
                   int64_t max_lag,
                   double[:] fsigma,
                   double[:] fdata,
                   double[:] data,
                   double[:] sigma2,
                   double[:] errors):
    cdef:
        int64_t j, t, loc

    for t in range(max_lag, nobs + burn):
        loc = 0
        fsigma[t] = parameters[loc]
        loc += 1
        for j in range(p):
            fsigma[t] += parameters[loc] * fdata[t - 1 - j]
            loc += 1
        for j in range(o):
            fsigma[t] += parameters[loc] * fdata[t - 1 - j] * (data[t - 1 - j] < 0)
            loc += 1
        for j in range(q):
            fsigma[t] += parameters[loc] * fsigma[t - 1 - j]
            loc += 1

        sigma2[t] = fsigma[t] ** (2.0 / power)
        data[t] = errors[t] * sqrt(sigma2[t])
        fdata[t] = abs(data[t]) ** power

    return data[burn:], sigma2[burn:]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def egarch_simulate(int64_t p,
                    int64_t o,
                    int64_t q,
                    int64_t nobs,
                    int64_t burn,
                    int64_t max_lag,
                    double[:] parameters,
                    double[:] data,
                    double[:] sigma2,
                    double[:] lnsigma2,
                    double[:] errors):
    cdef:
        int64_t j, t, loc
        double[:] abserrors = np.abs(errors)
        double norm_const = (2 / pi) ** 0.5

    for t in range(max_lag, nobs + burn):
        lnsigma2[t] = parameters[0]
        loc = 1
        for j in range(p):
            lnsigma2[t] += parameters[loc] * (abserrors[t - 1 - j] - norm_const)
            loc += 1
        for j in range(o):
            lnsigma2[t] += parameters[loc] * errors[t - 1 - j]
            loc += 1
        for j in range(q):
            lnsigma2[t] += parameters[loc] * lnsigma2[t - 1 - j]
            loc += 1

    sigma2 = np.exp(lnsigma2)
    data = errors * np.sqrt(sigma2)

    return data[burn:], sigma2[burn:]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def figarch_simulate(int64_t nobs,
                     int64_t burn,
                     int64_t truncation,
                     double power,
                     double omega_tilde,
                     double[:] data,
                     double[:] fdata,
                     double[:] fsigma,
                     double[:] sigma2,
                     double[:] lam_rev,
                     double[:] errors):
    cdef int64_t i, t

    for t in range(truncation, truncation + nobs + burn):
        fsigma[t] = omega_tilde
        for i in range(t - truncation, t):  # dot product
            fsigma[t] += lam_rev[i] * fdata[i]
        # fsigma[t] = omega_tilde + lam_rev.dot(fdata[t - truncation:t])

        sigma2[t] = fsigma[t] ** (2.0 / power)
        data[t] = errors[t] * sqrt(sigma2[t])
        fdata[t] = abs(data[t]) ** power

    return data[truncation + burn:], sigma2[truncation + burn:]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def harch_simulate(int64_t nobs,
                   int64_t burn,
                   int64_t max_lag,
                   double[:] parameters,
                   double[:] data,
                   double[:] sigma2,
                   double[:] errors,
                   int64_t[:] lags):
    cdef:
        int64_t i, j, t
        double param

    for t in range(max_lag, nobs + burn):
        sigma2[t] = parameters[0]
        for i in range(len(lags)):
            param = parameters[1 + i] / lags[i]
            for j in range(lags[i]):
                sigma2[t] += param * data[t - 1 - j] ** 2.0
        data[t] = errors[t] * sqrt(sigma2[t])

    return data[burn:], sigma2[burn:]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def garch_simulate_mc(int64_t p,
                      int64_t o,
                      int64_t q,
                      double power,
                      int64_t reps,
                      int64_t nobs,
                      int64_t burn,
                      double[::1] parameters,
                      double initial_value,
                      double[:, ::1] errors):
    cdef:
        int64_t max_lag = max([p, o, q])
        int64_t turns = nobs + burn
        int64_t r, t
        double[:, ::1] sigma2 = np.zeros((turns, reps), dtype=np.float64)
        double[:, ::1] data = np.zeros((turns, reps), dtype=np.float64)
        double[:, ::1] fsigma = np.zeros((turns, reps), dtype=np.float64)
        double[:, ::1] fdata = np.zeros((turns, reps), dtype=np.float64)

    fsigma[:max_lag] = initial_value
    sigma2[:max_lag] = initial_value ** (2.0 / power)

    for r in range(reps):
        for t in range(max_lag):
            data[t, r] = errors[t, r] * sigma2[t, r] ** 0.5
            fdata[t, r] = abs(data[t, r]) ** power

    return np.asarray(
        _garch_simulate_mc(p, o, q, power, reps, nobs, burn, parameters, fsigma, sigma2, fdata, data, errors),
        dtype=np.float64
    )


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double[:, ::1] _garch_simulate_mc(int64_t p,
                                       int64_t o,
                                       int64_t q,
                                       double power,
                                       int64_t reps,
                                       int64_t nobs,
                                       int64_t burn,
                                       double[::1] parameters,
                                       double[:, ::1] fsigma,
                                       double[:, ::1] sigma2,
                                       double[:, ::1] fdata,
                                       double[:, ::1] data,
                                       double[:, ::1] errors):
    cdef:
        int64_t j, r, t, loc
        int64_t max_lag = max([p, o, q])

    for r in range(reps):
        for t in range(max_lag, nobs + burn):
            loc = 0
            fsigma[t, r] = parameters[loc]
            loc += 1
            for j in range(p):
                fsigma[t, r] += parameters[loc] * fdata[t - 1 - j, r]
                loc += 1
            for j in range(o):
                fsigma[t, r] += parameters[loc] * fdata[t - 1 - j, r] * (data[t - 1 - j, r] < 0)
                loc += 1
            for j in range(q):
                fsigma[t, r] += parameters[loc] * fsigma[t - 1 - j, r]
                loc += 1

            sigma2[t, r] = fsigma[t, r] ** (2.0 / power)
            data[t, r] = errors[t, r] * sqrt(sigma2[t, r])
            fdata[t, r] = abs(data[t, r]) ** power

    return data[burn:]



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def egarch_simulate_mc(int64_t p,
                       int64_t o,
                       int64_t q,
                       int64_t nobs,
                       int64_t burn,
                       int64_t reps,
                       double[::1] parameters,
                       double[:, ::1] sigma2,
                       double[:, ::1] lnsigma2,
                       double[:, ::1] errors):
    cdef:
        int64_t r, max_lag = max([p, o, q])
        double[:, ::1] abserrors = np.abs(errors)
        double[:, ::1] data

    for r in range(reps):
        _egarch_simulate_mc(p, o, q, max_lag, nobs + burn, parameters, abserrors[:, r], lnsigma2[:, r], errors[:, r])

    sigma2 = np.exp(lnsigma2)
    data = errors * np.sqrt(sigma2)
    return np.asarray(data[burn:])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef _egarch_simulate_mc(int64_t p,
                         int64_t o,
                         int64_t q,
                         int64_t start,
                         int64_t end,
                         double[:] parameters,
                         double[:] abserrors,
                         double[:] lnsigma2,
                         double[:] errors):
    cdef:
        int64_t j, t, loc
        double norm_const = (2 / pi) ** 0.5

    for t in range(start, end):
        lnsigma2[t] = parameters[0]
        loc = 1

        for j in range(p):
            lnsigma2[t] += parameters[loc] * (abserrors[t - 1 - j] - norm_const)
            loc += 1
        for j in range(o):
            lnsigma2[t] += parameters[loc] * errors[t - 1 - j]
            loc += 1
        for j in range(q):
            lnsigma2[t] += parameters[loc] * lnsigma2[t - 1 - j]
            loc += 1



import numpy as np

cimport cython
from libc.stdint cimport int64_t


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def simulate_harx(
        double[:] y,
        int64_t nobs,
        int64_t k_x,
        int64_t max_lag,
        bint constant,
        double[:, :] x,
        double[:] errors,
        double[:] params,
        int64_t[:, :] lags):

    cdef:
        int64_t t, i, j, ind, lag_start, lag_end
        double ar
        int64_t[:] lag

    for t in range(max_lag, nobs):
        ind = 0
        if constant:
            y[t] = params[ind]
            ind += 1

        for j in range(len(lags)):
            ar, lag = 0.0, lags[j]
            for i in range(t-lag[1], t-lag[0]):
                ar += y[i]
            ar /= (lag[1] - lag[0])

            y[t] += params[ind] * ar
            ind += 1

        for i in range(k_x):
            y[t] += params[ind] * x[t, i]
        y[t] += errors[t]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def simulate_harx_mc(
        double[:, :] y,
        int64_t nobs,
        int64_t reps,
        int64_t k_x,
        int64_t max_lag,
        bint constant,
        double[:, :] x,
        double[:, :] errors,
        double[:] params,
        int64_t[:, :] lags):

    cdef double[:, :] data = _simulate_harx_mc(y, nobs, reps, k_x, max_lag, constant, x, errors, params, lags)
    return np.asarray(data, np.float64)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double[:, :] _simulate_harx_mc(
    double[:, :] y,
    int64_t nobs,
    int64_t reps,
    int64_t k_x,
    int64_t max_lag,
    bint constant,
    double[:, :] x,
    double[:, :] errors,
    double[:] params,
    int64_t[:, :] lags
):
    cdef:
        int64_t i, j, r, t, ind
        int64_t[:] lag
        double ar

    for r in range(reps):
        for t in range(max_lag, nobs):
            ind = 0
            if constant:
                y[t, r] = params[ind]
                ind += 1

            for j in range(len(lags)):
                ar, lag = 0.0, lags[j]
                for i in range(t-lag[1], t-lag[0]):
                    ar += y[i, r]
                ar /= (lag[1] - lag[0])

                y[t, r] += params[ind] * ar
                ind += 1

            for i in range(k_x):
                y[t, r] += params[ind] * x[t, i]
            y[t, r] += errors[t, r]

    return y



import numpy as np
import scipy.optimize as opt

from ._calibrate_utils import get_data_shape, validate_target_mean, validate_target_sd


def calibrate_mean_and_sd(data: np.ndarray, mean: np.ndarray, sd: np.ndarray, time_unit: int):
    validate_target_mean(data, mean)
    validate_target_sd(data, sd)

    sol = get_solutions(data, time_unit, mean, sd)

    for i in range(data.shape[2]):
        if sol[i, 1] > 0 and np.isfinite(sol[i]).all():
            data[..., i] = data[..., i] * sol[i, 1] + sol[i, 0]

    return data


def get_solutions(data: np.ndarray, time_unit: int, mean: np.ndarray, sd: np.ndarray):
    asset_list = (AssetMeanAndSD(data[..., i], time_unit, m, s) for i, (m, s) in enumerate(zip(mean, sd)))

    return np.array([f() for f in asset_list])


class AssetMeanAndSD:
    def __init__(self, data: np.ndarray, time_unit, target_mean: float, target_sd: float):
        assert data.ndim == 2
        self.data = data
        self.time_unit = time_unit
        self.years, self.trials = get_data_shape(data, time_unit)
        self.target_mean = target_mean
        self.target_sd = target_sd

        self.best_guess = np.array([self.calc_best_guess_mean(), self.calc_best_guess_sd()])

    def __call__(self) -> float:
        return opt.root(self.annualized_moments, x0=self.best_guess).x

    def annualized_moments(self, x: np.ndarray):
        calibrated_data = self.data * x[1] + x[0]
        return self.annualized_mean(calibrated_data), self.annualized_sd(calibrated_data)

    def annualized_mean(self, data) -> float:
        d = (data + 1).prod(0)
        return (np.sign(d) * np.abs(d) ** (1 / self.years)).mean() - 1 - self.target_mean

    def annualized_sd(self, data) -> float:
        return ((data + 1).reshape(self.years, self.time_unit, self.trials).prod(1) - 1).std(1).mean() - self.target_sd

    def calc_best_guess_mean(self):
        space = 2 ** np.linspace(-15, 4, 20)
        space = np.sort([*-space, *space])
        f_space = np.array([self.annualized_mean(self.data + x) for x in space])

        return self._best_guess_common(space, f_space)

    def calc_best_guess_sd(self):
        space = 2 ** np.linspace(-15, 6, 30)
        f_space = np.array([self.annualized_sd(self.data * x) for x in space])

        return self._best_guess_common(space, f_space)

    @staticmethod
    def _best_guess_common(space: np.ndarray, f_space: np.ndarray):
        mask = (f_space[:-1] * f_space[1:]) <= 0

        if not np.any(mask):
            raise RuntimeError("Unable to find roots")

        index = np.argmin(np.abs(f_space[:-1] - f_space[1:])[mask])  # index with best root character
        return (space[:-1][mask][index] + space[1:][mask][index]) / 2



import functools
from abc import ABC
from typing import Optional

import numpy as np


def _format_simulator(_simulate):
    """Formats input and output of the _simulator method"""

    @functools.wraps(_simulate)
    def decorator(self, size: int, reps: Optional[int] = None):
        if reps is not None:
            if not isinstance(reps, int) or reps <= 0:
                raise ValueError('`reps` must be an integer greater than 0')

        value = _simulate(self, size, reps)

        if isinstance(value, (int, float)):
            value = np.array([value], value)
        return np.asarray(value)

    return decorator


class DistributionMixin(ABC):
    def __init__(self):
        self._custom_dist: np.ndarray = None

    def check_dist_size(self, size):
        """Checks that the size required for the simulation is within the generated custom dist bounds"""
        if self.custom_dist is None:
            raise ValueError(f'method should not be called when custom_dist is not specified')

        if len(self.custom_dist) < size:
            times = round(np.ceil(size / len(self.custom_dist) * 100) / 100, 2)
            raise ValueError(f'rvs array too short. Increase the number of uniform rvs generated by {times}x')

    @property
    def custom_dist(self) -> Optional[np.ndarray]:
        """
        Optional density with fitted object from which to simulate

        The custom.dist option allows for defining a custom density which exists in the users workspace with methods for “r” (sampling, e.g. rnorm) and “d” (density e.g. dnorm). It must take a single fit object as its second argument. Alternatively, custom.dist can take any name in the name slot (e.g.“sample”) and a matrix in the fit slot with dimensions equal to m.sim (columns) and n.sim (rows). It is understood that what is supplied are the standardized (0,1) innovations and not the unstandardized residuals. The usefulness of this becomes apparent when one is considering the copula-GARCH approach or the bootstrap method.

        Returns
        -------
        {ndarray, None}
            Array of density values or None
        """
        return self._custom_dist

    @custom_dist.setter
    def custom_dist(self, values: Optional[np.ndarray]):
        if values is None:
            self._custom_dist = None
        else:
            values = np.asarray(values)
            self._custom_dist = float(values) if values.size == 1 else values



from .ged import GeneralizedError
from .normal import Normal
from .skew_student import SkewStudent
from .student import StudentsT

__all__ = ['GeneralizedError', 'Normal', 'SkewStudent', 'StudentsT']



from typing import Iterable

from statsmodels.iolib.summary import Summary as S


class Summary(S):
    def __init__(self, smry: S):
        self.header = None
        self.tables = smry.tables
        self.extra_txt = smry.extra_txt

    def add_header(self, header):
        self.header = header

    def as_text(self):
        """return tables as string

        Returns
        -------
        txt : string
            summary tables and extra text as one string

        """
        txt = super().as_text()
        if self.header:
            txt = f'{self.header}\n\n{txt}'
        return txt

    def as_latex(self):
        """return tables as string

        Returns
        -------
        latex : string
            summary tables and extra text as string of Latex

        Notes
        -----
        This currently merges tables with different number of columns.
        It is recommended to use `as_latex_tabular` directly on the individual
        tables.

        """
        latex = super().as_latex()
        if self.header:
            latex = f'{self.header}\n\n{latex}'
        return latex

    def as_csv(self):
        """return tables as string

        Returns
        -------
        csv : string
            concatenated summary tables in comma delimited format

        """
        csv = super().as_csv()
        if self.header:
            csv = f'{self.header}\n\n{csv}'
        return csv

    def as_html(self):
        """return tables as string

        Returns
        -------
        html : string
            concatenated summary tables in HTML format

        """
        html = super().as_html()
        if self.header:
            html = f'<h2>{self.header}</h2>{html}'
        return html


class SummaryList:
    def __init__(self, summaries: Iterable[Summary] = ()):
        self._summaries = list(summaries)

    @staticmethod
    def _assert_is_summary(obj):
        if not isinstance(obj, Summary):
            raise ValueError('object does not subclass Summary')

    def __getitem__(self, i):
        return self._summaries[i]

    def __setitem__(self, i, smry: Summary):
        self._assert_is_summary(smry)
        self._summaries[i] = smry

    def __str__(self):
        return self.as_text()

    def __repr__(self):
        # return '<' + str(type(self)) + '>\n"""\n' + self.__str__() + '\n"""'
        return str(type(self)) + '\n"""\n' + self.__str__() + '\n"""'

    def _repr_html_(self):
        """Display as HTML in IPython notebook."""
        return self.as_html()

    def append(self, smry: Summary):
        self._assert_is_summary(smry)
        self._summaries.append(smry)

    def extend(self, summaries: Iterable[Summary]):
        for s in summaries:
            self._assert_is_summary(s)
        self._summaries.extend(summaries)

    def as_text(self):
        """return tables as string

        Returns
        -------
        txt : string
            summary tables and extra text as one string

        """
        sep = '\n\n' + '*' * 100 + '\n\n'
        return sep.join(s.as_text() for s in self)

    def as_latex(self):
        """return tables as string

        Returns
        -------
        latex : string
            summary tables and extra text as string of Latex

        Notes
        -----
        This currently merges tables with different number of columns.
        It is recommended to use `as_latex_tabular` directly on the individual
        tables.

        """
        sep = '\n\n' + '*' * 100 + '\n\n'
        return sep.join(s.as_latex() for s in self)

    def as_csv(self):
        """return tables as string

        Returns
        -------
        csv : string
            concatenated summary tables in comma delimited format

        """
        sep = '\n\n' + '*' * 100 + '\n\n'
        return sep.join(s.as_csv() for s in self)

    def as_html(self):
        """return tables as string

        Returns
        -------
        html : string
            concatenated summary tables in HTML format

        """

        sep = '<br/>' * 2 + '<hr/>' + '<br/>' * 2
        return sep.join(s.as_html() for s in self)



from typing import List, Optional, Union

import numpy as np
import pandas as pd

from ._base import PARAMS, RNG_GEN, _ArchBase
from .distributions import GeneralizedError, Normal, SkewStudent, StudentsT
from .exceptions import NotFittedError
from .mean import ARX, ConstantMean, HARX, LS, ZeroMean
from .summary import Summary
from .volatility import ARCH, ConstantVariance, EGARCH, FIGARCH, GARCH, HARCH

__all__ = ['UArch']


class UArch(_ArchBase):
    """
    Univariate ARCH model that wraps on top of Mean, Volatility and Distribution classes defined in the `arch` package.
    Mainly, this class combines the original model and fitted model in the `arch` package for convenience. It has also
    some additional methods such as :code:`simulate_mc` for Monte Carlo simulations.
    """

    def __init__(self, mean='Constant', lags=0, vol='GARCH', p=1, o=0, q=1, power=2.0, dist='Normal', hold_back=None,
                 scale=1):
        """
        Creates the wrapping arch model

        Parameters
        ----------
        mean: { 'zero', 'constant', 'harx', 'har', 'ar', 'arx', 'ls' }, optional
            Name of the mean model.  Currently supported options are:

            * **Constant** (default) - Constant mean model
            * **Zero** - Zero mean model
            * **AR** - Autoregression model
            * **ARX** - Autoregression model with exogenous regressors. Falls back to **AR** if no exogenous regressors
            * **HAR** - Heterogeneous Autoregression model
            * **HARX** - Heterogeneous Autoregressions with exogenous regressors
            * **LS** - Least squares model

            For more information on the different models, check out the documentation at
            https://arch.readthedocs.io/en/latest/univariate/mean.html

        lags: int or list (int), optional
            Either a scalar integer value indicating lag length or a list of integers specifying lag locations.

        vol: { 'GARCH', 'ARCH', 'CONSTANT' 'EGARCH', 'FIGARCH' and 'HARCH', 'CONSTANT' }, optional
            Name of the volatility model.  Currently supported options are:

             * **GARCH** (default) - Standard GARCH process which can be used to specify the following models:

                * ARCH(p)
                * GARCH(p,q)
                * GJR-GARCH(p,o,q)
                * AVARCH(p)
                * AVGARCH(p,q)
                * TARCH(p,o,q)
                * Models with arbitrary, pre-specified powers

            * **ARCH** - ARCH process
            * **EGARCH** - EGARCH process
            * **FIGARCH** - Fractionally Integrated (FI) GARCH process
            * **HARCH** - Heterogeneous ARCH process
            * **Constant** (default) - Constant volatility process

        p: int, optional
            Lag order of the symmetric innovation

        o: int, optional
            Lag order of the asymmetric innovation

        q: int, optional
            Lag order of lagged volatility or equivalent

        power: float, optional
            Power to use with the innovations. Default is 2.0, which produces ARCH and related models. Using 1.0
            produces AVARCH and related models. Other powers can be specified, although these should be strictly
            positive, and usually larger than 0.25.

        dist: { 'normal', 'gaussian', 'studentst', 't', 'skewstudent', 'skewt', 'ged', 'generalized error' }, optional
            Name of the distribution for the innovations.  Currently supported options are:

            * **normal**, **gaussian** (default) - Standard Normal distribution
            * **t**, **studentst** - Standardized Student’s distribution
            * **skewstudent**, **skewt** - Standardized Skewed Student’s distribution.
            * **ged**, **generalized error" - Generalized Error Distribution

        hold_back: int, optional
            Number of observations at the start of the sample to exclude when estimating model parameters. Used when
            comparing models with different lag lengths to estimate on the common sample.

        scale: float
            Factor to scale data up or down by. This is useful when your data is too small leading to numerical errors
            when fitting. It will be used to scale simulation data
        """

        super(UArch, self).__init__()

        known_mean = ('zero', 'constant', 'harx', 'har', 'ar', 'arx', 'ls')
        known_vol = ('arch', 'figarch', 'garch', 'harch', 'constant', 'egarch')
        known_dist = ('normal', 'gaussian', 'studentst', 't', 'skewstudent',
                      'skewt', 'ged', 'generalized error')

        mean = mean.lower()
        vol = vol.lower()
        dist = dist.lower()
        if mean not in known_mean:
            raise ValueError(f"Unknown model type '{mean}' in mean")
        if vol.lower() not in known_vol:
            raise ValueError(f"Unknown model type '{vol}' in vol")
        if dist.lower() not in known_dist:
            raise ValueError(f"Unknown model type '{dist}' in dist")

        self._scale = scale

        self._model_setup = {
            'mean': mean,
            'lags': lags,
            'vol': vol,
            'p': p,
            'o': o,
            'q': q,
            'power': power,
            'dist': dist,
            'hold_back': hold_back
        }
        self._form_model(None)

    def fit(self, y, x=None, update_freq=1, disp='off', starting_values=None, cov_type='robust', show_warning=True,
            first_obs=None, last_obs=None, tol=None, options=None, backcast=None):
        r"""
        Fits the model given a nobs by 1 vector of sigma2 values

        Parameters
        ----------
        y: {ndarray, Series}
            The dependent variable

        x: {ndarray, DataFrame}, optional
            Exogenous regressors.  Ignored if model does not permit exogenous regressors.

        update_freq: int, optional
            Frequency of iteration updates.  Output is generated every `update_freq` iterations. Set to 0 to disable
            iterative output

        disp: 'final' or 'off' (default)
            Either 'final' to print optimization result or 'off' to display nothing

        starting_values: ndarray, optional
            Array of starting values to use.  If not provided, starting values are constructed by the model components

        cov_type:  str, optional
            Estimation method of parameter covariance.  Supported options are 'robust', which does not assume the
            Information Matrix Equality holds and 'classic' which does.  In the ARCH literature, 'robust' corresponds
            to Bollerslev-Wooldridge covariance estimator.

        show_warning: bool, optional
            Flag indicating whether convergence warnings should be shown.

        first_obs:  {int, str, datetime, Timestamp}
            First observation to use when estimating model

        last_obs: {int, str, datetime, Timestamp}
            Last observation to use when estimating model

        tol: float, optional
            Tolerance for termination

        options: dict, optional
            Options to pass to `scipy.optimize.minimize`.  Valid entries include 'ftol', 'eps', 'disp', and 'maxiter'

        backcast: float, optional
            Value to use as backcast. Should be measure :math:`\sigma^2_0`  since model-specific non-linear
            transformations are applied to value before computing the variance recursions.

        Returns
        -------
        UArch
            Fitted UArch instance
        """
        self._form_model(y, x)

        self._fit_model = self._model.fit(update_freq, disp, starting_values, cov_type, show_warning, first_obs,
                                          last_obs, tol, options, backcast)
        return self

    def forecast(self, params=None, horizon=1, start=None, align='origin', method='analytic',
                 simulations=1000, rng=None):
        """
        Construct forecasts from estimated model

        Parameters
        ----------
        params : ndarray, optional
            Alternative parameters to use.  If not provided, the parameters estimated when fitting the model are used.
            Must be identical in shape to the parameters computed by fitting the model.

        horizon : int, optional
           Number of steps to forecast

        start : {int, datetime, Timestamp, str}, optional
            An integer, datetime or str indicating the first observation to produce the forecast for. Datetimes can
            only be used with pandas inputs that have a datetime index. Strings must be convertible to a date time,
            such as in '1945-01-01'.

        align : {'origin', 'target'}, optional
            When set to 'origin', the t-th row of forecasts contains the forecasts for t+1, t+2, ..., t+h.

            When set to 'target', the t-th row contains the 1-step ahead forecast from time t-1, the 2 step from
            time t-2, ..., and the h-step from time t-h. 'target' simplifies computing forecast errors since the
            realization and h-step forecast are aligned.

        method : {'analytic', 'simulation', 'bootstrap'}, optional
            Method to use when producing the forecast. The default is 'analytic'. The method only affects the
            variance forecast generation.  Not all volatility models support all methods. In particular, volatility
            models that do not evolve in squares such as EGARCH or TARCH do not support the 'analytic' method
            for horizons > 1.

        simulations : int, optional
            Number of simulations to run when computing the forecast using either simulation or bootstrap.

        rng : {callable, ndarray}, optional
            If using a custom random number generator to for simulation-based forecasts, function must produce random
            samples using the syntax `rng(size)` where size is a 2-element tuple (simulations, horizon).

            Else, if a numpy array is passed in, array must have shape (simulation x horizon).

        Returns
        -------
        forecasts : ARCHModelForecast
            t by h data frame containing the forecasts.  The alignment of the forecasts is controlled by `align`.

        Notes
        -----
        The most basic 1-step ahead forecast will return a vector with the same length as the original data, where
        the t-th value will be the time-t forecast for time t + 1.  When the horizon is > 1, and when using the
        default value for `align`, the forecast value in position [t, h] is the time-t, h+1 step ahead forecast.

        If model contains exogenous variables (`model.x is not None`), then only 1-step ahead forecasts are available.
        Using horizon > 1 will produce a warning and all columns, except the first, will be nan-filled.

        If `align` is 'origin', forecast[t,h] contains the forecast made using y[:t] (that is, up to but not
        including t) for horizon h + 1.  For example, y[100,2] contains the 3-step ahead forecast using the first
        100 data points, which will correspond to the realization y[100 + 2]. If `align` is 'target', then the same
        forecast is in location [102, 2], so that it is aligned with the observation to use when evaluating,
        but still in the same column.
        """
        if params is None:
            params = self.params
        else:
            if params.size != np.array(self.params).size or params.ndim != self.params.ndim:
                raise ValueError('`params` have incorrect dimensions')

        if rng is not None and not callable(rng):
            rng = np.asarray(rng)
            if rng.shape != (simulations, horizon):
                raise ValueError(f"`rng` array should have dimensions ({(simulations, horizon)})")
            rng = lambda _: rng

        return self._fit_model.forecast(params, horizon, start, align, method, simulations, rng)

    def hedgehog_plot(self, params=None, horizon=10, step=10, start=None, type_='volatility', method='analytic',
                      simulations=1000):
        """
        Plot forecasts from estimated model

        Parameters
        ----------
        params: {Series, ndarray}, optional
            Alternative parameters to use.  If not provided, the parameters
            computed by fitting the model are used.  Must be 1-d and identical
            in shape to the parameters computed by fitting the model

        horizon: int, optional
            Number of steps to forecast

        step: int, optional
            Non-negative number of forecasts to skip between spines

        start: int, datetime or str, optional
            An integer, datetime or str indicating the first observation to produce the forecast for. Datetimes can only
            be used with pandas inputs that have a datetime index. Strings must be convertible to a date time, such as
            in '1945-01-01'. If not provided, the start is set to the earliest forecastable date

        type_: {'volatility', 'mean'}
            Quantity to plot, the forecast volatility or the forecast mean

        method: {'analytic', 'simulation', 'bootstrap'}
            Method to use when producing the forecast. The default is analytic. The method only affects the variance
            forecast generation.  Not all volatility models support all methods. In particular, volatility models that
            do not evolve in squares such as EGARCH or TARCH do not support the 'analytic' method for horizons > 1

        simulations: int
            Number of simulations to run when computing the forecast using either simulation or bootstrap

        Returns
        -------
        figure
            Handle to the figure
        """
        if params is None:
            params = self.params
        return self._fit_model.hedgehog_plot(params, horizon, step, start, type_, method, simulations)

    @property
    def params(self) -> pd.Series:
        """Model Parameters"""
        self._ensure_is_fitted()
        return self._fit_model.params

    def residual_plot(self, annualize=None, scale=None):
        """
        Plot standardized residuals and conditional volatility

        Parameters
        ----------
        annualize: str, optional
            String containing frequency of data that indicates plot should contain annualized volatility. Supported
            values are 'D' (daily), 'W' (weekly) and 'M' (monthly), which scale variance by 252, 52, and 12 respectively

        scale: float, optional
            Value to use when scaling returns to annualize.  If scale is provides, annualize is ignored and the value
            in scale is used.

        Returns
        -------
        figure
            Handle to the figure
        """
        return self._fit_model.plot(annualize, scale)

    def residuals(self, standardize=True) -> np.ndarray:
        """
        Model residuals

        Parameters
        ----------
        standardize: bool, optional
            Whether to standardize residuals. Residuals are standardized by dividing it with the conditional volatility

        Returns
        -------
        ndarray
            Residuals
        """
        model = self._fit_model
        lags = self._model_setup['lags']
        r = np.asarray(model.resid) / self._scale

        if standardize:
            r /= model.conditional_volatility
        return r[lags:]

    def simulate(self,
                 nobs: int,
                 burn=500,
                 initial_value: Union[np.ndarray, float] = None,
                 x: Optional[Union[np.ndarray, pd.DataFrame]] = None,
                 initial_value_vol: Union[np.ndarray, float] = None,
                 data_only=False,
                 params: PARAMS = None,
                 custom_dist: Optional[Union[RNG_GEN, np.ndarray]] = None) -> Union[pd.DataFrame, np.ndarray]:
        """
        Simulates data from a ARMA-GARCH model

        Parameters
        ----------
        nobs: int
            Length of series to simulate

        burn: int, optional
            Number of values to simulate to initialize the model and remove dependence on initial values

        initial_value: {ndarray, float}, optional
            Either a scalar value or `max(lags)` array set of initial values to use when initializing the model.
            If omitted, 0.0 is used

        x: {ndarray, DataFrame}, optional
            nobs + burn by k array of exogenous variables to include in the simulation. This should be a 2D matrix

        initial_value_vol: {ndarray, float}, optional
            An array or scalar to use when initializing the volatility process.

        data_only: bool, default True
            If True, this returns only the simulated data, omits the volatility and error. In this case, it will return
            as a numpy array. Otherwise, it returns a data frame with the data, volatility and error

        params: ndarray, optional
            If not None, model will use the parameters supplied to generate simulations. Otherwise, it will use the
            fitted parameters.

        custom_dist: {ndarray, Callable}, optional
            Optional density from which to simulate the innovations (Distribution) in the GARCH models. This is useful
            when working with the copula-GARCH model where each univariate model innovations has dependence on others.
            It is assumed that the values supplied are standardized [0, 1] innovations instead of the unstandardized
            residuals.

            The shape of the array must be at least as long as the simulation size required after accounting for burn
            and type of innovation process. If unsure, use :code:`simulation_size_required` to check.

            If a random number generator function is passed in, ensure that it only takes only argument and returns
            a numpy array. The argument can be an integer or a tuple of integers. In this case, the size will be
            automatically derived to save the user the trouble.

        Returns
        -------
        DataFrame or ndarray
            DataFrame with columns data containing the simulated values, volatility, containing the conditional
            volatility and errors containing the errors used in the simulation.
            If data_only, it returns the 'data' column as a numpy array
        """
        self._set_custom_dist(custom_dist, nobs, burn)
        if params is None:
            params = self.params

        return self._model.simulate(params, nobs, burn, initial_value, x, initial_value_vol, data_only) / self._scale

    def simulate_mc(self,
                    nobs: int,
                    reps: int,
                    burn=500,
                    initial_value: Union[np.ndarray, float] = None,
                    x: Optional[Union[np.ndarray, pd.DataFrame]] = None,
                    initial_value_vol: Union[np.ndarray, float] = None,
                    params: PARAMS = None,
                    custom_dist: Optional[Union[RNG_GEN, np.ndarray]] = None) -> Union[pd.DataFrame, np.ndarray]:
        """
        Simulates data from a ARMA-GARCH model with multiple repetitions.

        This is used for Monte Carlo simulations.

        Parameters
        ----------
        nobs: int
            Length of series to simulate

        reps: int
            Number of repetitions in Monte Carlo simulation

        burn: int, optional
            Number of values to simulate to initialize the model and remove dependence on initial values

        initial_value: {ndarray, float}, optional
            Either a scalar value or `max(lags)` array set of initial values to use when initializing the model.
            If omitted, 0.0 is used

        x: {ndarray, DataFrame}, optional
            nobs + burn by k array of exogenous variables to include in the simulation. This should be a 2D matrix

        initial_value_vol: {ndarray, float}, optional
            An array or scalar to use when initializing the volatility process.

        params: {Series, ndarray}, optional
            If not None, model will use the parameters supplied to generate simulations. Otherwise, it will use the
            fitted parameters.

        custom_dist: {ndarray, Callable}, optional
            Optional density from which to simulate the innovations (Distribution) in the GARCH models. This is useful
            when working with the copula-GARCH model where each univariate model innovations has dependence on others.
            It is assumed that the values supplied are standardized [0, 1] innovations instead of the unstandardized
            residuals.

            The shape of the array must be at least as long as the simulation size required after accounting for burn
            and type of innovation process. If unsure, use :code:`simulation_size_required` to check.

            If a random number generator function is passed in, he size will be automatically derived to save the user
            the trouble. However, the function must:

            * take as it first argument an integer or a tuple of integer
            * have other parameters that are optional
            * return a numpy array

        Returns
        -------
        simulated_data: ndarray
            Array containing the simulated values

        See Also
        --------
        UArch.simulation_horizon_required: Calculates the simulation size required
        """
        if not isinstance(reps, int) or reps < 1:
            raise ValueError('reps must be an integer greater than 0')

        self._set_custom_dist(custom_dist, nobs, burn, reps)
        if params is None:
            params = self.params

        return self._model.simulate_mc(params, nobs, reps, burn, initial_value, x, initial_value_vol) / self._scale

    def summary(self, short=False, dp=4) -> Union[pd.Series, Summary]:
        """
        Summary of fitted model

        Parameters
        ----------
        short: bool, optional
            Whether to show short summary or full summary.

        dp: int, optional
            Number of decimal places to show in short summary

        Returns
        -------
        Summary
            Model Summary
        """
        self._ensure_is_fitted()
        if short:
            params = self._fit_model.params.round(dp)
            se = self._fit_model.std_err.round(dp)
            return pd.Series(['{p:.{dp}f} ± {s:.{dp}f}'.format(p=p, s=se[i], dp=dp) for i, p in enumerate(params)],
                             index=params.index)
        else:
            return Summary(self._fit_model.summary())

    def _ensure_is_fitted(self):
        """Ensures model is fitted. Else raises error"""
        if self._fit_model is None:
            raise NotFittedError(f"This ArchModel instance is not fitted yet")

    def _form_model(self, y, x=None):
        """
        Forms the actual ARCH model. This is a convenience method so that users can specify data at run time
        instead of at class instantiation
        """
        mean: str = self._model_setup['mean']
        lags: Union[int, List[int]] = self._model_setup['lags']
        vol: str = self._model_setup['vol']
        p: int = self._model_setup['p']
        o: int = self._model_setup['o']
        q: int = self._model_setup['q']
        power: float = self._model_setup['power']
        dist: str = self._model_setup['dist']
        hold_back: Optional[int] = self._model_setup['hold_back']

        y = y * self._scale if y is not None else None
        if x is not None:
            x = np.asarray(x) * self._scale
            if x.ndim == 1:
                x = x.reshape(-1, 1)

        if mean == 'zero':
            am = ZeroMean(y, hold_back=hold_back)
        elif mean == 'constant':
            am = ConstantMean(y, hold_back=hold_back)

        elif mean == 'arx':
            am = ARX(y, x, lags, hold_back=hold_back)
        elif mean == 'ar':
            am = ARX(y, None, lags, hold_back=hold_back)
        elif mean == 'harx':
            am = HARX(y, x, lags, hold_back=hold_back)
        elif mean == 'har':
            am = HARX(y, None, lags, hold_back=hold_back)
        else:  # mean == 'ls'
            am = LS(y, x, hold_back=hold_back)

        if vol == 'constant':
            v = ConstantVariance()
        elif vol == 'arch':
            v = ARCH(p=p)
        elif vol == 'figarch':
            v = FIGARCH(p=p, q=q)
        elif vol == 'garch':
            v = GARCH(p=p, o=o, q=q, power=power)
        elif vol == 'egarch':
            v = EGARCH(p=p, o=o, q=q)
        else:  # vol == 'harch'
            v = HARCH(lags=p)

        if dist in ('skewstudent', 'skewt'):
            d = SkewStudent()
        elif dist in ('studentst', 't'):
            d = StudentsT()
        elif dist in ('ged', 'generalized error'):
            d = GeneralizedError()
        else:  # ('gaussian', 'normal')
            d = Normal()

        am.volatility = v
        am.distribution = d
        self._model = am

    def __model_description__(self, include_lags=True, include_fitted_stats=False):
        """Generates the model description for use by __str__ and related functions"""
        desc = self._model._model_description(include_lags)

        if self._fit_model is not None and include_fitted_stats:
            pass

        return desc

    def _repr_html_(self):
        """HTML representation for IPython Notebook"""
        desc = self.__model_description__(include_fitted_stats=True)

        html = f"<strong>{self._model.name}</strong>"

        fragment = ",\n".join(f"<strong>{key}: </strong> {val}" for key, val in desc.items())
        html = f"{html}(\n{fragment}\n<strong>ID: </strong>: {hex(id(self))})"

        return html

    def _set_custom_dist(self, custom_dist, nobs, burn, reps=None):
        horizon = self.simulation_horizon_required(nobs, burn)
        if isinstance(reps, int):  # for monte carlo
            size = horizon, reps
        else:
            size = horizon

        if custom_dist is not None:
            if callable(custom_dist):
                self._model.distribution.custom_dist = custom_dist(size)
            else:
                custom_dist = np.asarray(custom_dist)
                if len(custom_dist) < horizon:
                    raise ValueError(f"`custom_dist` array size should at least be {horizon} long")
                if reps is not None and reps != custom_dist.shape[1]:
                    raise ValueError(f"number of columns `custom_dist` array should equals {reps}")

                self._model.distribution.custom_dist = custom_dist

    def __repr__(self):
        txt = self.__str__()
        return txt + f', {hex(id(self))}'

    def __str__(self):
        desc = self.__model_description__()
        return f"{self._model.name}({', '.join(f'{k}: {v}' for k, v in desc.items() if k and v)})"



from typing import Tuple

import numpy as np

__all__ = ["calc_current_sd", "get_data_shape", "validate_target_mean", "validate_target_sd"]


def calc_current_sd(data: np.ndarray, time_unit: int):
    years, trials, num_assets = get_data_shape(data, time_unit)
    return ((data + 1).reshape(years, time_unit, trials, num_assets).prod(1) - 1).std(1).mean(0)


def get_data_shape(data: np.ndarray, time_unit: int):
    def get_years(periods):
        years = periods / time_unit
        assert years.is_integer(), "time_unit must divide data periods to an integer year"
        return int(years)

    def get_3d_shape() -> Tuple[int, int, int]:
        periods, trials, num_assets = data.shape
        return get_years(periods), trials, num_assets

    def get_2d_shape() -> Tuple[int, int]:
        periods, trials = data.shape
        return get_years(periods), trials

    if data.ndim == 2:
        return get_2d_shape()
    else:
        return get_3d_shape()


def validate_target_mean(data: np.ndarray, mean: np.ndarray):
    assert len(mean) == data.shape[2], "length of target returns (mean) must equal number of assets"
    assert np.isfinite(np.asarray(mean)).all(), "all target returns (mean) must be finite"


def validate_target_sd(data: np.ndarray, sd: np.ndarray):
    _, _, num_assets = data.shape
    assert num_assets == len(sd), "length of target vols (sd) must equal number of assets"
    assert np.isfinite(np.asarray(sd)).all(), "all target vols (sd) must be finite"
    assert (np.asarray(sd) >= 0).all(), "target vol must all be >= 0"



from copy import deepcopy

import numpy as np


def truncate_outliers(data: np.ndarray, sd=0, replacement='mean', inplace=False):
    """
    Truncates outliers by replacing it with the mean, median or a specified value.

    Outlier is determined by the number of standard deviations past the mean within the asset group.

    Parameters
    ----------
    data: ndarray
        The tensor (data cube) where the axis represents time, trials and number of asset classes respectively

    sd: float
        The number of standard deviations to consider a point an outlier. If :code:`sd` is set to 0, no changes will
        be made

    replacement: {float, 'mean', 'median'}
        The value to replace outliers with. Valid values are 'mean', 'median' or a number.

    inplace: bool
        If True, calibration will modify the original data. Otherwise, a deep copy of the original data will be
        made before calibration. Deep copy can be time consuming if data is big.

    Returns
    -------
    ndarray
        A data cube with the outliers replaced
    """

    assert sd >= 0, "Standard deviations to determine outliers must be >= 0"
    if isinstance(replacement, str):
        assert replacement.lower() in ('mean', 'median'), "replacement can only be 'mean', 'median' or a float value"
    else:
        assert isinstance(replacement, float), "replacement can only be 'mean', 'median' or a float value"

    if not inplace:
        data = deepcopy(data)

    if sd == 0:
        return data

    def get_replacement(returns_matrix):
        if replacement == "mean":
            return np.mean(returns_matrix)
        elif replacement == "median":
            return np.median(returns_matrix)
        else:
            return float(replacement)

    for i in range(data.shape[2]):
        returns = data[..., i]
        mean, std = returns.mean(), returns.std()
        r = get_replacement(returns)

        returns[returns > (mean + sd * std)] = r
        returns[returns < (mean - sd * std)] = r

    return data



from typing import Union

__all__ = ['get_integer_time_unit']


def get_integer_time_unit(time_unit: Union[int, str]):
    if isinstance(time_unit, int):
        return time_unit
    if isinstance(time_unit, str):
        if time_unit.lower() in ('week', 'weekly'):
            return 52
        if time_unit.lower() in ('month', 'monthly'):
            return 12
        if time_unit.lower() in ('quarter', 'quarterly'):
            return 4
        if time_unit.lower() in ('semi-annual', 'semi-annually'):
            return 2
        if time_unit.lower() in ('annual', 'annually', 'year', 'yearly'):
            return 1

    raise ValueError(f"Unacceptable time_unit argument, {time_unit}")



from typing import Optional

from arch.univariate.distribution import Normal as N
from scipy.stats import norm

from ._base import DistributionMixin, _format_simulator


class Normal(N, DistributionMixin):
    def __init__(self, random_state=None):
        DistributionMixin.__init__(self)
        N.__init__(self, random_state)

    @_format_simulator
    def _simulator(self, size: int, reps: Optional[int] = None):
        if self.custom_dist is None:
            if reps is not None:
                size = size, reps
            return self._random_state.standard_normal(size)

        else:
            self.check_dist_size(size)
            ppf = norm.ppf(self.custom_dist[:size])
            self.custom_dist = None  # reset simulator

            return ppf



from functools import wraps

import numpy as np
import pandas as pd
from arch.utility.array import ensure1d

from ._mean_simulations import simulate_harx, simulate_harx_mc

__all__ = ['harx_simulator', 'harx_mc_simulator']


def harx_simulator(class_func):
    class_func.__docs__ = """
        Simulates data from a linear regression, AR or HAR models
        
        Parameters
        ----------
        params: ndarray
            Parameters to use when simulating the model. Parameter order is
            [mean volatility distribution] where the parameters of the mean
            model are ordered [constant lag[0] lag[1] ... lag[p] ex[0] ...
            ex[k-1]] where lag[j] indicates the coefficient on the jth lag in
            the model and ex[j] is the coefficient on the jth exogenous
            variable.
            
        nobs: int
            Length of series to simulate

        burn: int, default 500
            Number of values to simulate to initialize the model and remove dependence on initial values

        initial_value: {ndarray, float}, optional
            Either a scalar value or `max(lags)` array set of initial values to use when initializing the model.
            If omitted, 0.0 is used

        x: {ndarray, DataFrame}, optional
            nobs + burn by k array of exogenous variables to include in the simulation. This should be a 2D matrix

        initial_value_vol: {ndarray, float}, optional
            An array or scalar to use when initializing the volatility process.

        data_only: bool, default True
            If True, this returns only the simulated data, omits the volatility and error. In this case, it will return
            as a numpy array. Otherwise, it returns a data frame with the data, volatility and error

        Returns
        -------
        simulated_data : Union[DataFrame, numpy array]
            DataFrame with columns data containing the simulated values, volatility, containing the conditional
            volatility and errors containing the errors used in the simulation.
            If data_only, it returns the 'data' column as a numpy array
        
        Examples
        --------
        >>> import numpy as np
        >>> from muarch import UArch
        >>> from muarch.volatility import GARCH
        >>> harx = UArch(lags=[1, 5, 22])
        >>> harx.volatility = GARCH()
        >>> harx_params = np.array([1, 0.2, 0.3, 0.4])
        >>> garch_params = np.array([0.01, 0.07, 0.92])
        >>> params = np.concatenate((harx_params, garch_params))
        >>> sim_data = harx.simulate(params, 1000)

        Simulating models with exogenous regressors requires the regressors
        to have nobs plus burn data points

        >>> nobs = 100
        >>> burn = 200
        >>> x = np.random.randn(nobs + burn, 2)
        >>> x_params = np.array([1.0, 2.0])
        >>> params = np.concatenate((harx_params, x_params, garch_params))
        >>> sim_data = harx.simulate(params, nobs=nobs, burn=burn, x=x)
    """

    @wraps(class_func)
    def decorator(model, params, nobs, burn=500, initial_value=None, x=None, initial_value_vol=None, data_only=False):
        k_x = 0
        if x is not None:
            x = np.asarray(x)
            k_x = x.shape[1]
            if x.shape[0] != nobs + burn:
                raise ValueError('x must have nobs + burn rows')

        # added model._lags is not None for LS models
        lags = model._lags if model._lags is not None else np.zeros((0, 0), int)

        mc = int(model.constant) + lags.shape[1] + k_x
        vc = model.volatility.num_params
        dc = model.distribution.num_params
        num_params = mc + vc + dc
        params = ensure1d(params, 'params', series=False)
        if params.shape[0] != num_params:
            raise ValueError(f'params has the wrong number of elements. Expected {num_params}, got {params.shape[0]}. '
                             f'Perhaps you forgot to add the exogenous variables?')

        dist_params = [] if dc == 0 else params[-dc:]
        vol_params = params[mc:mc + vc]
        simulator = model.distribution.simulate(dist_params)
        sim_data = model.volatility.simulate(vol_params,
                                             nobs + burn,
                                             simulator,
                                             burn,
                                             initial_value_vol)
        errors = sim_data[0]
        vol = np.sqrt(sim_data[1])

        max_lag = np.max(lags) if lags.size else 0

        if initial_value is None:
            initial_value = 0.0
        elif not np.isscalar(initial_value):
            initial_value = ensure1d(initial_value, 'initial_value')
            if initial_value.shape[0] != max_lag:
                raise ValueError('initial_value has the wrong shape')

        y = np.zeros(nobs + burn)
        y[:max_lag] = initial_value

        simulate_harx(y, nobs + burn, k_x, max_lag, model.constant, x, errors, params, lags.T.astype(np.int64))

        if data_only:
            return y[burn:]

        df = dict(data=y[burn:], volatility=vol[burn:], errors=errors[burn:])
        df = pd.DataFrame(df)
        return df

    return decorator


def harx_mc_simulator(class_func):
    class_func.__docs__ = """
        Monte Carlo simulation from a linear regression, AR or HAR models
        
        Parameters
        ----------
        params: ndarray
            Parameters to use when simulating the model. Parameter order is
            [mean volatility distribution] where the parameters of the mean
            model are ordered [constant lag[0] lag[1] ... lag[p] ex[0] ...
            ex[k-1]] where lag[j] indicates the coefficient on the jth lag in
            the model and ex[j] is the coefficient on the jth exogenous
            variable.

        nobs: int
            Length of series to simulate

        reps: int
            Number of trials in the Monte Carlo simulation

        burn: int, default 500
            Number of values to simulate to initialize the model and remove dependence on initial values

        initial_value: {ndarray, float}, optional
            Either a scalar value or `max(lags)` array set of initial values to use when initializing the model.
            If omitted, 0.0 is used

        x: {ndarray, DataFrame}, optional
            nobs + burn by k array of exogenous variables to include in the simulation. This should be a 2D matrix

        initial_value_vol: {ndarray, float}, optional
            An array or scalar to use when initializing the volatility process.

        Returns
        -------
        simulated_data : ndarray
            matrix of simulated values

        Examples
        --------
        >>> import numpy as np
        >>> from muarch import MUArch
        >>> AR1_GARCH11= MUArch(3, mean='ARX', lags=1, vol='GARCH', p=1, q=1, scale=100)

        >>> sim_data = AR1_GARCH11.simulate_mc(nobs=36, reps=1000, n_jobs=-1)
        """

    @wraps(class_func)
    def decorator(model, params, nobs, reps, burn=500, initial_value=None, x=None, initial_value_vol=None):
        k_x = 0
        if x is not None:
            x = np.asarray(x)
            k_x = x.shape[1]
            if x.shape[0] != nobs + burn:
                raise ValueError('x must have nobs + burn rows')

        # added model._lags is not None for LS models
        lags = model._lags if model._lags is not None else np.zeros((0, 0), int)

        mc = int(model.constant) + lags.shape[1] + k_x
        vc = model.volatility.num_params
        dc = model.distribution.num_params
        num_params = mc + vc + dc
        params = ensure1d(params, 'params', series=False)
        if params.shape[0] != num_params:
            raise ValueError(f'params has the wrong number of elements. Expected {num_params}, got {params.shape[0]}. '
                             f'Perhaps you forgot to add the exogenous variables?')

        dist_params = [] if dc == 0 else params[-dc:]
        vol_params = params[mc:mc + vc]
        simulator = model.distribution.simulate(dist_params)
        errors = model.volatility.simulate_mc(vol_params, nobs + burn, reps, simulator, burn, initial_value_vol)

        max_lag = np.max(lags) if lags.size else 0

        if initial_value is None:
            initial_value = 0
        elif np.isscalar(initial_value):
            initial_value = np.repeat(initial_value, max_lag)
        elif not np.isscalar(initial_value):
            initial_value = ensure1d(initial_value, 'initial_value')
            if initial_value.shape[0] != max_lag:
                raise ValueError('initial_value has the wrong shape')

        y = np.zeros((nobs + burn, reps), np.float64)  # y is modified in place
        y[:max_lag, :reps] = initial_value
        return simulate_harx_mc(y, nobs + burn, reps, k_x, max_lag, model.constant, x, errors, params,
                                lags.T.astype(np.int64))[burn:]

    return decorator



import os

from setuptools import Extension, find_packages, setup

import versioneer

try:
    from Cython.Build import cythonize
    from Cython.Compiler import Options
    from Cython.Distutils.build_ext import new_build_ext as build_ext

    USE_CYTHON = True
except ImportError:
    from distutils.command.build_ext import build_ext

    USE_CYTHON = False


    def cythonize(x, *args, **kwargs):
        return x


    class Options:
        pass

DEV_BUILD = os.environ.get('DEV_BUILD', '0').lower() in ('true', '1')
PACKAGE_NAME = 'muarch'

cmdclass = versioneer.get_cmdclass()
cmdclass.update({
    'build_ext': build_ext
})

with open('README.md') as f:
    long_description = f.read()


def build_ext_modules():
    macros = [('NPY_NO_DEPRECATED_API', '1'),
              ('NPY_1_7_API_VERSION', '1')]

    modules = [
        {
            'name': 'muarch.volatility._vol_simulations',
            'sources': ['muarch/volatility/_vol_simulations.pyx'],
        },
        {
            'name': 'muarch.mean._mean_simulations',
            'sources': ['muarch/mean/_mean_simulations.pyx'],
        }
    ]

    extensions = []
    for m in modules:
        # if not built with Cython, use the c or cpp files
        language = m.get('language', 'c')
        ext = '.pyx' if USE_CYTHON else f'.{language}'

        for i, source in enumerate(m['sources']):
            _file, _ext = os.path.splitext(source)
            m['sources'][i] = _file + ext

        extensions.append(Extension(**m, language=language, define_macros=macros))

    # compiler options
    Options.annotate = DEV_BUILD

    # compiler directives
    directives = {'language_level': '3', 'profile': DEV_BUILD, 'linetrace': DEV_BUILD}

    return cythonize(extensions, compiler_directives=directives)


setup_requires = [
    'cython >=0.29',
    'numpy >=1.15'
]

install_requires = [
    'arch >=4.7',
    'copulae >=0.4.0',
    'numpy >=1.15',
    'scipy >=1.1',
    'setuptools >=40.8',
    'pandas >=0.23'
]

setup(
    name=PACKAGE_NAME,
    license='MIT',
    version=versioneer.get_version().split('+')[0],
    description='Multiple Univariate ARCH modeling toolbox built on top of the ARCH package',
    author='Daniel Bok',
    author_email='daniel.bok@outlook.com',
    packages=find_packages(include=['muarch', 'muarch.*']),
    ext_modules=build_ext_modules(),
    url='https://github.com/DanielBok/muarch',
    cmdclass=cmdclass,
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Cython',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
    ],
    install_requires=install_requires,
    setup_requires=setup_requires,
    python_requires='>=3.6',
    include_package_data=True,
    zip_safe=False
)



from warnings import warn

import numpy as np
from arch.univariate.volatility import (
    ARCH as A,
    ConstantVariance as C,
    EGARCH as E,
    FIGARCH as F,
    GARCH as G,
    HARCH as H
)
from arch.utility.exceptions import InitialValueWarning, initial_value_warning

from muarch.volatility import _vol_simulations as vs
from .simulators import RNG, garch_simulator, garch_simulator_mc

try:
    from arch.univariate.recursions import figarch_weights
except ImportError:
    from arch.univariate.recursions_python import figarch_weights


class ARCH(A):
    @garch_simulator
    def simulate(self, parameters: np.ndarray, nobs: int, rng: RNG, burn=500, initial_value=None):
        pass

    @garch_simulator_mc
    def simulate_mc(self, parameters: np.ndarray, nobs: int, reps: int, rng: RNG, burn=500, initial_value=None):
        pass


class ConstantVariance(C):
    def simulate(self, parameters, nobs, rng, burn=500, initial_value=None):
        errors = rng(nobs + burn)
        sigma2 = np.ones(nobs + burn) * parameters[0]
        data = np.sqrt(sigma2) * errors
        return data[burn:], sigma2[burn:]

    def simulate_mc(self, parameters: np.ndarray, nobs: int, reps: int, rng: RNG, burn=500, initial_value=None):
        errors = rng(nobs + burn, reps)
        sigma2 = np.ones((nobs + burn, reps)) * parameters[0]
        data = np.sqrt(sigma2) * errors
        return data[burn:]


class EGARCH(E):
    def simulate(self, parameters, nobs, rng, burn=500, initial_value=None):
        p, o, q = self.p, self.o, self.q
        errors = rng(nobs + burn)
        max_lag = max(p, o, q)

        sigma2, data, lnsigma2 = self._simulation_setup(nobs + burn, parameters, initial_value, errors)

        return vs.egarch_simulate(p, o, q, nobs, burn, max_lag, parameters, data, sigma2, lnsigma2, errors)

    def simulate_mc(self, parameters: np.ndarray, nobs: int, reps: int, rng: RNG, burn=500, initial_value=None):
        p, o, q = self.p, self.o, self.q
        errors = rng(nobs + burn, reps)

        sigma2, data, lnsigma2 = self._simulation_setup((nobs + burn, reps), parameters, initial_value, errors)

        return vs.egarch_simulate_mc(p, o, q, nobs, burn, reps, parameters, sigma2, lnsigma2, errors)

    def _simulation_setup(self, shape, parameters, initial_value, errors):
        p, o, q = self.p, self.o, self.q
        max_lag = max(p, o, q)

        if initial_value is None:
            if q > 0:
                beta_sum = np.sum(parameters[p + o + 1:])
            else:
                beta_sum = 0.0

            if beta_sum < 1:
                initial_value = parameters[0] / (1.0 - beta_sum)
            else:
                warn(initial_value_warning, InitialValueWarning)
                initial_value = parameters[0]

        sigma2 = np.zeros(shape)
        data = np.zeros(shape)
        lnsigma2 = np.zeros(shape)

        lnsigma2[:max_lag] = initial_value
        sigma2[:max_lag] = np.exp(initial_value)
        data[:max_lag] = errors[:max_lag] * np.sqrt(sigma2[:max_lag])

        return sigma2, data, lnsigma2


class FIGARCH(F):
    def simulate(self, parameters, nobs, rng, burn=500, initial_value=None):
        truncation = self.truncation
        p, q, power = self.p, self.q, self.power
        lam = figarch_weights(parameters[1:], p, q, truncation)
        lam_rev = lam[::-1]
        errors = rng(truncation + nobs + burn)

        if initial_value is None:
            persistence = np.sum(lam)
            beta = parameters[-1] if q else 0.0

            initial_value = parameters[0]
            if beta < 1:
                initial_value /= (1 - beta)
            if persistence < 1:
                initial_value /= (1 - persistence)
            if persistence >= 1.0 or beta >= 1.0:
                warn(initial_value_warning, InitialValueWarning)

        sigma2 = np.empty(truncation + nobs + burn)
        data = np.empty(truncation + nobs + burn)
        fsigma = np.empty(truncation + nobs + burn)
        fdata = np.empty(truncation + nobs + burn)

        fsigma[:truncation] = initial_value
        sigma2[:truncation] = initial_value ** (2.0 / power)
        data[:truncation] = np.sqrt(sigma2[:truncation]) * errors[:truncation]
        fdata[:truncation] = abs(data[:truncation]) ** power
        omega = parameters[0]
        beta = parameters[-1] if q else 0
        omega_tilde = omega / (1 - beta)

        return vs.figarch_simulate(nobs, burn, truncation, power, omega_tilde, data, fdata, fsigma, sigma2, lam_rev,
                                   errors)

    def simulate_mc(self, parameters: np.ndarray, reps: int, nobs: int, rng: RNG, burn=500, initial_value=None):
        raise NotImplementedError


class GARCH(G):
    @garch_simulator
    def simulate(self, parameters: np.ndarray, nobs: int, rng: RNG, burn=500, initial_value=None):
        pass

    @garch_simulator_mc
    def simulate_mc(self, parameters: np.ndarray, reps: int, nobs: int, rng: RNG, burn=500, initial_value=None):
        pass


class HARCH(H):
    def simulate(self, parameters, nobs, rng, burn=500, initial_value=None):
        lags = self.lags
        errors = rng(nobs + burn)

        if initial_value is None:
            if (1.0 - np.sum(parameters[1:])) > 0:
                initial_value = parameters[0] / (1.0 - np.sum(parameters[1:]))
            else:
                warn(initial_value_warning, InitialValueWarning)
                initial_value = parameters[0]

        sigma2 = np.empty(nobs + burn)
        data = np.empty(nobs + burn)
        max_lag = np.max(lags)
        sigma2[:max_lag] = initial_value
        data[:max_lag] = np.sqrt(initial_value)

        return vs.harch_simulate(nobs, burn, max_lag, parameters, data, sigma2, errors, lags.astype(np.int64))

    def simulate_mc(self, parameters: np.ndarray, reps: int, nobs: int, rng: RNG, burn=500, initial_value=None):
        raise NotImplementedError



from typing import Optional

import numpy as np
from arch.univariate.distribution import SkewStudent as SS
from scipy.stats import t, uniform

from ._base import DistributionMixin, _format_simulator


class SkewStudent(DistributionMixin, SS):
    def __init__(self, random_state=None):
        DistributionMixin.__init__(self)
        SS.__init__(self, random_state)

    @_format_simulator
    def _simulator(self, size: int, reps: Optional[int] = None) -> np.ndarray:
        nu = self._parameters[0]  # degrees of freedom
        if self.custom_dist is None:
            if reps is not None:
                size = size, reps
            return self.ppf(uniform.rvs(size=size), nu)
        else:
            self.check_dist_size(size)
            ppf = self.ppf(self.custom_dist[:size], nu)
            self.custom_dist = None  # reset simulator

            return ppf

    def ppf(self, pits, parameters=None):
        self._check_constraints(parameters)

        scalar = np.isscalar(pits)
        if scalar:
            pits = np.array([pits])
        eta, lam = parameters

        a = self.__const_a(parameters)
        b = self.__const_b(parameters)

        cond = pits < (1 - lam) / 2

        # slight speed up for really large problems
        icdf1 = t._ppf(pits[cond] / (1 - lam), eta)
        icdf2 = t._ppf(.5 + (pits[~cond] - (1 - lam) / 2) / (1 + lam), eta)
        icdf = -999.99 * np.ones_like(pits)
        icdf[cond] = icdf1
        icdf[~cond] = icdf2
        icdf = (icdf *
                (1 + np.sign(pits - (1 - lam) / 2) * lam) * (1 - 2 / eta) ** .5 -
                a)
        icdf = icdf / b

        if scalar:
            icdf = icdf[0]
        return icdf



from functools import wraps
from typing import Callable, Optional, Union
from warnings import warn

import numpy as np
from arch.utility.exceptions import InitialValueWarning, initial_value_warning

from ._vol_simulations import garch_simulate, garch_simulate_mc

RNG = Union[Callable[[int], np.ndarray], Callable[[int, Optional[int]], np.ndarray]]


def garch_simulator(class_func):
    @wraps(class_func)
    def decorator(model, parameters: np.ndarray, nobs: int, rng: RNG, burn=500, initial_value=None):
        p, o, q, power = model.p, model.o, model.q, model.power
        errors = rng(nobs + burn)

        if initial_value is None:
            scale = np.ones_like(parameters)
            scale[p + 1:p + o + 1] = 0.5

            persistence = np.sum(parameters[1:] * scale[1:])
            if (1.0 - persistence) > 0:
                initial_value = parameters[0] / (1.0 - persistence)
            else:
                warn(initial_value_warning, InitialValueWarning)
                initial_value = parameters[0]

        sigma2 = np.zeros(nobs + burn)
        data = np.zeros(nobs + burn)
        fsigma = np.zeros(nobs + burn)
        fdata = np.zeros(nobs + burn)

        max_lag = np.max([p, o, q])
        fsigma[:max_lag] = initial_value
        sigma2[:max_lag] = initial_value ** (2.0 / power)
        data[:max_lag] = np.sqrt(sigma2[:max_lag]) * errors[:max_lag]
        fdata[:max_lag] = abs(data[:max_lag]) ** power

        return garch_simulate(p, o, q, power, parameters, nobs, burn, max_lag, fsigma, fdata, data, sigma2, errors)

    return decorator


def garch_simulator_mc(class_func):
    @wraps(class_func)
    def decorator(model, parameters: np.ndarray, nobs: int, reps: int, rng: RNG, burn=500, initial_value=None):
        p, o, q, power = model.p, model.o, model.q, model.power
        errors = rng(nobs + burn, reps)

        if initial_value is None:
            scale = np.ones_like(parameters)
            scale[p + 1:p + o + 1] = 0.5

            persistence = np.sum(parameters[1:] * scale[1:])
            if (1.0 - persistence) > 0:
                initial_value = parameters[0] / (1.0 - persistence)
            else:
                initial_value = parameters[0]

        return garch_simulate_mc(p, o, q, power, reps, nobs, burn, parameters, initial_value, errors)

    return decorator



from pathlib import Path

import pandas as pd

__all__ = ['load_etf']

data_dir: Path = Path(__file__).parent.joinpath('data')


def load_etf(typ='returns') -> pd.DataFrame:
    """
    Monthly returns or price from 2010-09-01 to 2019-03-01 of

    * Vanguard S&P 500 ETF (VOO)
    * iShares MSCI Emerging Markets ETF
    * Vanguard Total World Stock ETF

    By default, function gives returns data, which has one last day of data (first day is NA).

    Parameters
    ----------
    typ: {'returns', 'price'}
        Determines whether price or returns data is required

    Returns
    -------
    DataFrame
        DataFrame containing monthly price or returns data with the date as the index
    """
    fp = data_dir.joinpath('etf.csv')
    df = pd.read_csv(fp, sep=r'\s*,\s*', parse_dates=['Date'], engine='python', index_col=[0])

    if typ == 'price':
        return df
    else:
        return df.pct_change().dropna()
